from __future__ import annotations
from collections import defaultdict
from dataclasses import dataclass
import dataclasses
import datetime
from decimal import Decimal as PyDecimal
from typing import Optional, Union, Tuple
from v0.relationalai import debugging
from v0.relationalai.semantics.metamodel import builtins, helpers, ir, types, visitor, compiler, factory as f
from v0.relationalai.semantics.metamodel.util import OrderedSet, ordered_set
import rich
import sys

#--------------------------------------------------
# Helpers
#--------------------------------------------------

def to_relation(node:ir.Lookup|ir.Update|ir.Aggregate) -> ir.Relation:
    """ Get the relation being referred to by this node. """
    if isinstance(node, ir.Aggregate):
        return node.aggregation
    return node.relation

def to_name(type:ir.Type) -> str:
    """ Get a human-readable name representing a type. """
    if isinstance(type, ir.DecimalType):
        return f"Decimal({type.precision},{type.scale})"
    elif isinstance(type, ir.ScalarType):
        return type.name
    elif isinstance(type, ir.UnionType):
        return '|'.join([to_name(t) for t in type.types])
    elif isinstance(type, ir.ListType):
        return f"[{to_name(type.element_type)}]"
    else:
        raise TypeError(f"Unknown type: {type}")

def check_int64(value: int):
    INT64_MIN = -(2 ** 63)
    INT64_MAX = 2 ** 63 - 1

    if not (INT64_MIN <= value <= INT64_MAX):
        raise OverflowError(f"Value '{value}' does not fit in Int64.")

#--------------------------------------------------
# Constants
#--------------------------------------------------

# map potential type conversions to their cost
# The differences in the costs of converting from one type to another should be
# smaller than the cost of a literal conversion.
# Otherwise, we'll introduce spurious conversions of say Int->Decimal64 plus a cheap literal conversion
# Float->Decimal64, rather than a single literal conversion Int->Float.
# Use a large enough number here so that if all arguments are literals, it's cheaper to convert
# the literals (cost 10) rather than not convert some literals and convert a non-literal instead.
# For instance, for `y=1+x` is should be cheaper to convert 1 to float and not convert x and y, than it would be
# convert both x and y to integer.
# We do however, want to prefer converting to number types lower in the lattice.
# So we make the cost depend only on the result type.
# Making all the costs the same would introduce spurious unions of conversions.

CONVERSION_LATTICE:dict[tuple[ir.Type, ir.Type], int] = {
    # Make the cost depend only on the result type.
    (types.Int64, types.Int128): 1000,
    (types.Int64, types.GenericDecimal): 1002,
    (types.Int64, types.Float): 1004,
    (types.Int128, types.GenericDecimal): 1002,
    (types.GenericDecimal, types.Float): 1003,
    (types.Int128, types.Float): 1004,
    # Int -> Hash (UInt128)
    (types.Int64, types.Hash): 1001,
    (types.Int128, types.Hash): 1001,
    (types.Int128, types.UInt128): 1001,
}

def conversion_lattice_cost(from_type, to_type):
    """
    Lookup the conversion cost in the conversion lattice to convert from_type to to_type.
    Return inf is the conversion is not allowed.
    """
     # the conversion lattice only has generic decimals as placeholders
    if isinstance(from_type, ir.DecimalType):
        from_type = types.GenericDecimal
    if isinstance(to_type, ir.DecimalType):
        to_type = types.GenericDecimal

    if (from_type, to_type) in CONVERSION_LATTICE:
        return CONVERSION_LATTICE[(from_type, to_type)]
    return float("inf")

# list of non-parametric types that are primitives. Avoid using this constant, use
# is_base_primitive instead because it accounts for parametric types like decimals.
_NON_PARAMETRIC_PRIMITIVES = [
    types.Int64,
    types.Int128,
    types.UInt128,
    types.Float,
    types.GenericDecimal,
    types.Bool,
    types.String,
    types.Date,
    types.DateTime,
    types.Hash
]

ENTITY_LIKES = [types.RowId]

# these are relations that try to preserve their input types
# e.g. if you add a USD and a Decimal, the result is still a USD
TYPE_PRESERVERS = [
    *builtins.plus.overloads,
    *builtins.minus.overloads,
    *builtins.mul.overloads,
    *builtins.mod.overloads,
    builtins.trunc_div,
    builtins.concat,
    *builtins.abs.overloads,
    *builtins.sum.overloads,
    *builtins.max.overloads,
    *builtins.min.overloads,
    # Exclude div overloads (like Int/Int -> Float) where the result type is different from the input types
    *(filter(lambda x: x.fields[0].type == x.fields[1].type == x.fields[2].type, builtins.div.overloads)),
    # Exclude avg overloads (avg Int -> Float) where the result type is different from the input types
    *(filter(lambda x: x.fields[0].type == x.fields[1].type, builtins.avg.overloads)),
]

#--------------------------------------------------
# Type helpers
#--------------------------------------------------

def is_abstract(type:ir.Type) -> bool:
    return types.is_abstract_type(type)

def to_type(value: ir.Value|ir.Field) -> ir.Type:
    if isinstance(value, (ir.Var, ir.Field, ir.Literal)):
        return value.type

    if isinstance(value, tuple):
        return types.AnyList

    if isinstance(value, ir.Relation):
        return types.AnyList

    raise TypeError(f"Cannot determine IR type for value: {value} of type {type(value).__name__}")

def type_matches(actual:ir.Type, expected:ir.Type, allow_expected_parents=False) -> bool:
    """
    True iff we can use a value of the actual type when expecting the expected type.

    TODO: document allow_expected_parents (useful for checking that a Person is an Adult)
    """
    # exact match
    if actual == expected:
        return True

    # any matches anything
    if actual == types.Any or expected == types.Any:
        return True

    # any entity matches any entity (surprise surprise!)
    if extends_any_entity(expected) and not is_primitive(actual):
        return True

    # all decimals match across each other
    if types.is_decimal(actual) and types.is_decimal(expected):
        return True

    # scalar matches with super-type handling
    is_scalar = isinstance(actual, ir.ScalarType)
    exp_is_scalar = isinstance(expected, ir.ScalarType)
    if is_scalar and any([type_matches(parent, expected) for parent in actual.super_types]):
        return True
    if allow_expected_parents and exp_is_scalar and any([type_matches(actual, parent, allow_expected_parents) for parent in expected.super_types]):
        return True
    if not is_primitive(expected) and exp_is_scalar and actual in ENTITY_LIKES:
        return True

    # a union type matches if any of its types match the expected type
    if isinstance(actual, ir.UnionType):
        for t in actual.types:
            if type_matches(t, expected):
                return True

    # types that match Number
    if actual == types.Number and types.is_number(expected):
        return True
    if expected == types.Number and types.is_number(actual):
        return True
    return False


def conversion_allowed(arg:ir.Value, actual:ir.Type, expected:ir.Type, arg_types:set[ir.Type]=set()) -> bool:
    return conversion_cost(arg, actual, expected, arg_types) < float("inf")

def literal_conversion_allowed(actual:ir.Type, expected:ir.Type) -> bool:
    if actual in (types.Int64, types.Int128) and (expected == types.Hash or types.is_number(expected)):
        return True
    elif types.is_decimal(actual) and expected == types.Float:
        return True
    elif types.is_decimal(actual) and types.is_decimal(expected):
        return True
    return False

def conversion_cost(arg: ir.Value, actual:ir.Type, expected:ir.Type, arg_types:set[ir.Type]) -> float:
    if type_matches(actual, expected):
        return 0

    # if we have a type variable and all the types in the expression
    # match, then there's no promotion needed and this is a valid overload
    if expected is types.EntityTypeVar and not is_primitive(actual) and len(arg_types) == 1:
        return 0

    # Literal numbers can be converted to other number types (but not to value types).
    if isinstance(arg, ir.Literal):
        if literal_conversion_allowed(actual, expected):
            return 50

    # Value types that don't match
    if is_value_type(actual) and is_value_type(expected):
        return float("inf")

    base_actual = to_base_primitive(actual)
    base_expected = to_base_primitive(expected)

    if not base_actual or not base_expected:
        return float("inf")

    if type_matches(base_actual, base_expected):
        return 0

    return conversion_lattice_cost(base_actual, base_expected)


def merge_types(type1: ir.Type, type2: ir.Type, is_literal1:bool, is_literal2:bool) -> ir.Type:
    if type1 == type2:
        return type1

    # give precedence to nominal types (e.g. merging USD(decimal) with decimal gives USD(decimal))
    base_primitive_type1 = to_base_primitive(type1)
    base_primitive_type2 = to_base_primitive(type2)
    if base_primitive_type1 == base_primitive_type2:
        if is_base_primitive(type1):
            return type2
        elif is_base_primitive(type2):
            return type1

    combined = ordered_set()
    types_to_process = [type1, type2]

    # Iterative flattening of union types
    while types_to_process:
        t = types_to_process.pop()
        if isinstance(t, ir.UnionType):
            types_to_process.extend(t.types)
        else:
            combined.add(t)

    # If we have multiple types and Any is one of them, remove Any
    if len(combined) > 1 and types.Any in combined:
        combined.remove(types.Any)

    # If we have multiple types and one is a literal and the other is not,
    # we can try to convert the literal to the other type.
    if len(combined) > 1:
        if is_literal1 and not is_literal2 and literal_conversion_allowed(type1, type2):
            return type2
        if is_literal2 and not is_literal1 and literal_conversion_allowed(type2, type1):
            return type1

    # Return single type or create a union
    return next(iter(combined)) if len(combined) == 1 else f.union_type(list(combined))

def to_base_primitive(type:ir.Type) -> Optional[ir.Type]:
    if isinstance(type, ir.ScalarType):
        if is_base_primitive(type):
            return type
        # walk the hierarchy to find a base primitive
        for parent in type.super_types:
            if found := to_base_primitive(parent):
                return found
    if isinstance(type, ir.UnionType):
        for t in type.types:
            if found := to_base_primitive(t):
                return found
    return None

def is_value_type(type:ir.Type):
    return isinstance(type, ir.ScalarType) and is_primitive(type) and not is_base_primitive(type)

def is_base_primitive(type:ir.Type) -> bool:
    return type in _NON_PARAMETRIC_PRIMITIVES or types.is_decimal(type)

def is_primitive(type:ir.Type) -> bool:
    return to_base_primitive(type) is not None

def extends_any_entity(type:ir.Type) -> bool:
    if type == types.AnyEntity:
        return True
    if isinstance(type, ir.ScalarType):
        for parent in type.super_types:
            if extends_any_entity(parent):
                return True
    return False

def invalid_type(type:ir.Type) -> bool:
    if isinstance(type, ir.UnionType):
        # if there are multiple primitives, or a primtive and a non-primitive
        # then we have an invalid type
        if len(type.types) > 1:
            return any([is_primitive(t) for t in type.types])
    return False

def try_preserve_type(types:set[ir.Type]) -> Optional[ir.Type]:
    # we keep the input type as the output type if either all inputs
    # are the exact same type or there's one nominal and its base primitive
    # type, e.g. USD + Decimal
    if len(types) == 1:
        return next(iter(types))
    if len(types) == 2:
        t1, t2 = types
        # type preservation is not applicable to decimal types
        if isinstance(t1, ir.DecimalType) and isinstance(t2, ir.DecimalType):
            return None
        base_equivalent = type_matches(t1, t2, allow_expected_parents=True)
        if base_equivalent:
            # as long as one of the types is a base primitive, we can use the
            # other type as final preserved type
            if is_base_primitive(t1):
                return t2
            elif is_base_primitive(t2):
                return t1
    return None

# Was this variable created for a shared literal?
def var_is_from_literal(var: ir.Var) -> bool:
    return var.name.startswith("__literal__")

#--------------------------------------------------
# Type Errors
#--------------------------------------------------

@dataclass
class TyperError():
    node: ir.Node

@dataclass
class TypeMismatch(TyperError):
    expected: ir.Type
    actual: ir.Type
    def __str__(self):
        return f"[red bold][Type Mismatch][/red bold] expected [yellow]{to_name(self.expected)}[/yellow], got [red]{to_name(self.actual)}[/red]: [white]{str(self.node).strip()}[/white]"

@dataclass
class InvalidType(TyperError):
    type: ir.Type
    def __str__(self):
        return f"[red bold][Invalid Type][/red bold] incompatible types [red]{to_name(self.type)}[/red]: [white]{str(self.node).strip()}[/white]"

@dataclass
class UnresolvedOverload(TyperError):
    arg_types: list[ir.Type]
    def __str__(self):
        assert isinstance(self.node, (ir.Lookup, ir.Update, ir.Aggregate))
        rel = to_relation(self.node)
        types = ', '.join([to_name(t) for t in self.arg_types])
        return f"[red bold][Unresolved Overload][/red bold] [yellow]{rel.name}({types})[/yellow]: [white]{str(self.node).strip()}[/white]"

@dataclass
class UnresolvedType(TyperError):
    def __str__(self):
        return f"[red bold][Unresolved Type][/red bold] [white]{str(self.node).strip()}[/white]"
    pass

class TyperContextEnricher(visitor.Visitor):
    def __init__(self):
        super().__init__()
        # store mapping from the concept population relation var id to a concept population type
        self.concept_population_types:dict[int, ir.Type] = {}

    def visit_lookup(self, node: ir.Lookup, parent: Optional[ir.Node]):
        rel = to_relation(node)
        if helpers.is_concept_lookup(rel):
            for arg, field in zip(node.args, rel.fields):
                if isinstance(arg, ir.Var):
                    arg_type = arg.type
                    field_type = field.type
                    # When the rule declares something like `Adult(v::Person)`, we must record that `v` is actually of type
                    #   `Adult` so this refined type is remembered for later checks.
                    if field_type != arg_type and type_matches(arg_type, field_type, True):
                        self.concept_population_types[arg.id] = field_type

#--------------------------------------------------
# Propagation Network
#--------------------------------------------------

# The core idea of the typer is to build a propagation network where nodes are vars, fields,
# or overloaded lookups/updates/aggregates. The intuition is that _all_ types in the IR
# ultimately flow from relation fields, so if we figure those out we just need to propagate
# their types to unknown vars, which may then flow into other fields and so on.
#
# This means the network only needs to contain nodes that either directly flow into
# an abstract node or are themselves abstract. We need to track overloads because
# their arguments effectively act like abstract vars until we've resolved the final types.

Node = Union[ir.Var, ir.Field, ir.Lookup, ir.Update, ir.Aggregate, ir.Literal]

class PropagationNetwork():
    def __init__(self):

        # track the set of nodes that represent entry points into the network
        self.roots = ordered_set()
        # we separately want to track nodes that were loaded from a previous run
        # so that even if we have edges to them, we _still_ consider them roots
        # and properly propagate types from them at the beginning
        self.loaded_roots = set()
        self.edges:dict[Node, dict[int, Node]] = defaultdict(dict)
        self.has_incoming = set()
        self.type_requirements:dict[Node, OrderedSet[ir.Field]] = defaultdict(lambda: ordered_set())

        self.errors:list[TyperError] = []
        # all fields for which types were resolved
        self.fields = set()
        self.resolved_types:dict[int, ir.Type] = {}
        # overloads resolved for a lookup/update/aggregate, by node id
        self.resolved_overloads:dict[int, list[ir.Relation]] = {}
        # if the node resolves to an overload the uses GenericDecimal, the concrete
        # DecimalType to be used (assumes there's a single overload with decimals for a node)
        self.resolved_overload_decimal:dict[int, ir.DecimalType] = {}
        self.node_is_literal:dict[int, bool] = {}

    #--------------------------------------------------
    # Mismatches
    #--------------------------------------------------

    def mismatch(self, node:Node, expected:ir.Type, actual:ir.Type):
        self.errors.append(TypeMismatch(node, expected, actual))

    def invalid_type(self, node:Node, type:ir.Type):
        self.errors.append(InvalidType(node, type))

    def unresolved_overload(self, node:ir.Lookup|ir.Update|ir.Aggregate):
        self.errors.append(UnresolvedOverload(node, [self.resolve(a) for a in node.args]))

    def unresolved_type(self, node:Node):
        self.errors.append(UnresolvedType(node))

    def has_errors(self, node:Node) -> bool:
        for mismatch in self.errors:
            if mismatch.node == node:
                return True
        return False

    #--------------------------------------------------
    # Types and Edges
    #--------------------------------------------------

    def add_edge(self, source:Node, target:Node):
        if target not in self.loaded_roots:
            self.roots.remove(target)
        if source not in self.has_incoming:
            self.roots.add(source)
        self.edges[source][target.id] = target
        self.has_incoming.add(target)

    def add_type(self, node:Node, type:ir.Type, is_literal:bool):
        if isinstance(node, ir.Field):
            self.fields.add(node)

        if node.id in self.resolved_types:
            if isinstance(node, ir.Literal) and self.resolved_types[node.id] == node.type:
                # if the previous "resolved type" was just the literal type itself, we can
                # replace it; this is safe because literals are only part of one task
                self.resolved_types[node.id] = type
            else:
                self.resolved_types[node.id] = merge_types(self.resolved_types[node.id], type, self.node_is_literal[node.id], is_literal)
        else:
            self.resolved_types[node.id] = type

        if node.id in self.node_is_literal:
            is_literal &= self.node_is_literal[node.id]
        self.node_is_literal[node.id] = is_literal

    def add_type_requirement(self, source:Node, field:ir.Field):
        self.type_requirements[source].add(field)

    #--------------------------------------------------
    # Load previous types
    #--------------------------------------------------

    def load_types(self, type_dict:dict[ir.Field, ir.Type]):
        for node, type in type_dict.items():
            self.add_type(node, type, False)
            self.loaded_roots.add(node)
            self.roots.add(node)

    #--------------------------------------------------
    # Resolve
    #--------------------------------------------------

    def resolve(self, value:Node|ir.Value) -> ir.Type:
        if isinstance(value, (ir.Var, ir.Field)):
            return self.resolved_types.get(value.id) or to_type(value)
        assert not isinstance(value, (ir.Lookup, ir.Update, ir.Aggregate)), "Should never try to resolve a task"
        return to_type(value)

    #--------------------------------------------------
    # Overloads
    #--------------------------------------------------

    def resolve_overload_deps(self, op:ir.Lookup|ir.Update|ir.Aggregate, keep_known=False):
        # We need to find which args flow into this task, which flow out, and
        # if some are inputs, then they are required before we can resolve
        incoming:list[Node] = []
        outgoing:list[Node] = []
        required:list[Node] = []
        relation = to_relation(op)
        field_args = zip(relation.fields, [self.resolve(a) for a in op.args], op.args)
        has_inputs = any([field.input for field in relation.fields])
        for field, arg_type, arg in field_args:
            if not isinstance(arg, (ir.Var, ir.Literal)):
                continue
            arg_is_abstract = is_abstract(arg_type)
            if not arg_is_abstract and not keep_known:
                continue
            elif field.input or not arg_is_abstract:
                incoming.append(arg)
                required.append(arg)
            elif not has_inputs:
                incoming.append(arg)
                outgoing.append(arg)
            else:
                outgoing.append(arg)

        # check if the overloads of this relation are fully resolved, if any aren't
        # then their fields are incoming edges to this node
        for overload in relation.overloads:
            for field in overload.fields:
                resolved_type = self.resolve(field)
                if resolved_type != types.GenericDecimal and is_abstract(resolved_type):
                    incoming.append(field)
                    required.append(field)

        return incoming, outgoing, required

    def resolve_overload(self, op:ir.Lookup|ir.Update|ir.Aggregate) -> Optional[list[ir.Relation]]:
        # check if we have any unresolved args that are required, if all of our args
        # are unresolved (len(incoming) == len(relation.fields)) then we have no information
        # to try and resolve the overloads with
        incoming, outgoing, required = self.resolve_overload_deps(op)
        relation = to_relation(op)
        if required or len(incoming) == len(relation.fields):
            return

        overloads = relation.overloads
        if not overloads:
            overloads = [relation]

        # otherwise we compute the cost of each overload and return the set of relations
        # that have the lowest cost. This can be multiple in e.g. the Person.pets.name
        # case where Cat_name and Dog_name are equally valid
        inf = float("inf")
        min_cost = inf
        matches = []
        for overload in overloads:
            arg_types = set(self.resolve(arg) for arg, field in zip(op.args, overload.fields) if field.input)
            total = 0
            for arg, field in zip(op.args, overload.fields):
                arg_type = self.resolve(arg)
                field_type = self.resolve(field)
                total += conversion_cost(arg, arg_type, field_type, arg_types)
                if total == inf:
                    break
            if total != inf and total <= min_cost:
                if total < min_cost:
                    min_cost = total
                    matches.clear()
                matches.append(overload)
        return matches

    #--------------------------------------------------
    # Propagation
    #--------------------------------------------------

    def propagate(self):
        edges = self.edges
        work_list = []

        # go through all the roots and find any that are not abstract, they'll
        # be the first nodes to push types through the network
        unhandled = ordered_set()
        for node in self.roots:
            if not isinstance(node, (ir.Var, ir.Field, ir.Literal)):
                continue
            node_type = self.resolve(node)
            if not is_abstract(node_type):
                is_literal = isinstance(node, ir.Literal) or (isinstance(node, ir.Var) and var_is_from_literal(node))
                self.add_type(node, node_type, is_literal)
                work_list.append(node)
            else:
                unhandled.add(node)

        # We need to visit nodes in topological order; that is we need to visit all predecessors
        # of a node before the node itself.

        # push known type nodes through the edges
        while work_list:
            source = work_list.pop(0)
            unhandled.remove(source)
            source_type = self.resolve(source)
            is_literal = self.node_is_literal.get(source.id, False)
            # check to see if the source has ended up with a set of types that
            # aren't valid, e.g. a union of primitives
            if invalid_type(source_type):
                self.invalid_type(source, source_type)

            # propagate our type to each outgoing edge
            for out in edges.get(source, {}).values():
                # if this is an overload then we need to try and resolve it
                if isinstance(out, (ir.Update, ir.Lookup, ir.Aggregate)):
                    found = self.resolve_overload(out)
                    if found is not None:
                        start_arg_types = [self.resolve(arg) for arg in out.args]
                        self.propagate_overloads(out, found)
                        final_arg_types = [self.resolve(arg) for arg in out.args]
                        for arg, start, final in zip(out.args, start_arg_types, final_arg_types):
                            if start != final:
                                work_list.append(arg)
                # otherwise, we just add to the outgoing node's type and if it
                # changes we add it to the work list
                elif start := self.resolve(out):
                    self.add_type(out, source_type, is_literal)
                    if start != self.resolve(out) or out in unhandled:
                        work_list.append(out)

        for source in unhandled:
            self.unresolved_type(source)

        # now that we've pushed all the types through the network, we need to validate
        # that all type requirements of those nodes are met
        for node, fields in self.type_requirements.items():
            node_type = self.resolve(node)
            for field in fields:
                field_type = self.resolve(field)
                if not type_matches(node_type, field_type):
                    node_base = to_base_primitive(node_type)
                    field_base = to_base_primitive(field_type)
                    if conversion_lattice_cost(node_base, field_base) == float("inf"):
                        self.mismatch(node, field_type, node_type)

        # return the resolved type for fields only
        field_types = {}
        for field in self.fields:
            if field.id in self.resolved_types:
                field_types[field] = self.resolved_types[field.id]
        return field_types


    def propagate_overloads(self, node:ir.Lookup|ir.Update|ir.Aggregate, overloads:list[ir.Relation]):
        if not overloads:
            return self.unresolved_overload(node)

        # we've resolved the overloads, so store that
        self.resolved_overloads[node.id] = overloads

        # we need to determine the final types of our args but taking all the overloads
        # and adding the type of their fields back to the args.
        resolved_args = [self.resolve(a) for a in node.args]
        for overload in overloads:
            resolved_fields = [self.resolve(f) for f in overload.fields]
            if types.GenericDecimal in resolved_fields:
                # this overload contains generic decimals, so find which specific type of
                # decimal to use given the arguments being passed
                decimal, resolved_fields = self.specialize_decimal_overload(resolved_fields, resolved_args)
                self.resolved_overload_decimal[node.id] = decimal

            # if our overload preserves types, we check to see if there's a preserved
            # output type given the inputs and if so, shadow the field's type with the
            # preserved type
            if overload in TYPE_PRESERVERS:
                input_types = set([arg_type for field, arg_type
                                    in zip(overload.fields, resolved_args)
                                    if field.input])
                if out_type := try_preserve_type(input_types):
                    resolved_fields = [field_type if field.input else out_type
                                        for field, field_type in zip(overload.fields, resolved_fields)]

            for field_type, arg_type, arg in zip(resolved_fields, resolved_args, node.args):
                if isinstance(arg, ir.Var) and is_abstract(arg_type):
                    is_literal = var_is_from_literal(arg)
                    self.add_type(arg, field_type, is_literal)
                elif isinstance(arg, ir.Literal) and arg_type != field_type:
                    self.add_type(arg, field_type, True)

        return None

    def specialize_decimal_overload(self, field_types:list[ir.Type], arg_types:list[ir.Type]) -> Tuple[ir.DecimalType, list[ir.Type]]:
        """
        Find the decimal type to use for an overload that has GenericDecimals in its field_types,
        and which is being referred to with these arg_types.

        Return a tuple where the first element is the specialized decimal type, and the second
        element is a new list that contains the same types as field_types but with
        GenericDecimal replaced by this specialized decimal.
        """
        decimal = None
        for arg_type in arg_types:
            x = types.decimal_supertype(arg_type)
            if isinstance(x, ir.DecimalType):
                # the current specialization policy is to select the decimal with largest
                # scale and, if there multiple with the largest scale, the one with the
                # largest precision. This is safe because when converting a decimal to the
                # specialized decimal, we never truncate fractional digits (because we
                # selected the largest scale) and, if the non-fractional digits are too
                # large to fit the specialized decimal, we will have a runtime overflow,
                # which should alert the user of the problem.
                #
                # In the future we can implement more complex policies. For example,
                # snowflake has well documented behavior for how the output of operations
                # behave in face of different decimal types, and we may use that:
                # https://docs.snowflake.com/en/sql-reference/operators-arithmetic#scale-and-precision-in-arithmetic-operations
                if decimal is None or x.scale > decimal.scale or (x.scale == decimal.scale and x.precision > decimal.precision):
                    decimal = x
        assert(isinstance(decimal, ir.DecimalType))
        return decimal, [decimal if field_type == types.GenericDecimal else field_type
                for field_type in field_types]


    #--------------------------------------------------
    # Display
    #--------------------------------------------------

    # draw the network as a mermaid graph for the debugger
    def to_mermaid(self, max_edges=500) -> str:
        resolved = self.resolved_types
        nodes = ordered_set()
        link_strs = []
        for src, dsts in self.edges.items():
            nodes.add(src)
            for dst in dsts.values():
                if len(link_strs) > max_edges:
                    break
                nodes.add(dst)
                link_strs.append(f"n{src.id} --> n{dst.id}")
            if len(link_strs) > max_edges:
                break

        for src, dsts in self.type_requirements.items():
            nodes.add(src)
            for dst in dsts:
                if len(link_strs) > max_edges:
                    break
                nodes.add(dst)
                link_strs.append(f"n{src.id} --> n{dst.id}")
            if len(link_strs) > max_edges:
                break

        def type_span(t:ir.Type) -> str:
            type_str = t.name if isinstance(t, ir.ScalarType) else str(t)
            return f"<span style='color:cyan;'>{type_str.strip()}</span>"

        def overload_span(rel:ir.Relation, arg_types:list[ir.Type]) -> str:
            args = []
            for field, arg_type in zip(rel.fields, arg_types):
                field_type = self.resolve(field)
                if not type_matches(arg_type, field_type) and field_type != types.EntityTypeVar:
                    args.append(f"<span style='color:yellow;'>{str(arg_type).strip()} -> {str(field_type).strip()}</span>")
                elif isinstance(arg_type, ir.UnionType):
                    args.append(type_span(field_type))
                else:
                    args.append(type_span(arg_type))
            return f'{rel.name}({", ".join(args)})'

        node_strs = []
        for node in nodes:
            klass = ""
            if isinstance(node, ir.Var):
                ir_type = resolved.get(node.id) or self.resolve(node)
                type_str = type_span(ir_type)
                # if this is one of our vars that is a placeholder for a literal, just use the type
                if isinstance(ir_type, ir.ScalarType) and var_is_from_literal(node):
                    label = f'(["Literal {type_str}"])'
                else:
                    label = f'(["{node.name}: {type_str}"])'
            elif isinstance(node, ir.Literal):
                ir_type = resolved.get(node.id) or self.resolve(node)
                type_str = type_span(ir_type)
                label = f'(["Literal {type_str}"])'
            elif isinstance(node, ir.Field):
                ir_type = resolved.get(node.id) or self.resolve(node)
                type_str = type_span(ir_type)
                klass = ":::field"
                label = f"{{{{\"{node.name}: {type_str}\"}}}}"
            elif isinstance(node, (ir.Lookup, ir.Update, ir.Aggregate)):
                arg_types = [self.resolve(arg) for arg in node.args]
                if node.id in self.resolved_overloads:
                    overloads = self.resolved_overloads[node.id]
                    content = "<br/>".join([overload_span(o, arg_types) for o in overloads])
                else:
                    content = overload_span(to_relation(node), arg_types)
                label = f'[/"{content}"/]'
            else:
                raise NotImplementedError(f"Unknown node type: {type(node)}")
            if self.has_errors(node):
                klass = ":::error"
            node_strs.append(f"n{node.id}{label}{klass}")

        node_str = "\n                ".join(node_strs)
        link_str = "\n                ".join(link_strs)
        template = f"""
            %%{{init: {{'theme':'dark', 'flowchart':{{'useMaxWidth':false, 'htmlLabels': true}}}}}}%%
            flowchart TD
                linkStyle default stroke:#666
                classDef field fill:#245,stroke:#478
                classDef error fill:#624,stroke:#945,color:#f9a
                classDef default stroke:#444,stroke-width:2px, font-size:12px

                %% nodes
                {node_str}

                %% edges
                {link_str}
        """
        return template

    # simplified, less verbose (compared to mermaid) output for snapshot testing
    def to_fish(self) -> str:
        resolved = self.resolved_types
        nodes = ordered_set()
        for src, dsts in self.edges.items():
            nodes.add(src)
            nodes.update(dsts.values())
        for src, dsts in self.type_requirements.items():
            nodes.add(src)
            nodes.update(dsts)

        def type_to_str(t:ir.Type) -> str:
            type_str = t.name if isinstance(t, ir.ScalarType) else str(t)
            return type_str.strip()

        def overload_str(rel:ir.Relation, arg_types:list[ir.Type]) -> str:
            args = []
            for field, arg_type in zip(rel.fields, arg_types):
                field_type = self.resolve(field)
                if not type_matches(arg_type, field_type):
                    args.append(f"{str(arg_type).strip()} -?> {str(field_type).strip()}")
                elif isinstance(arg_type, ir.UnionType):
                    args.append(type_to_str(field_type))
                else:
                    args.append(type_to_str(arg_type))
            return f'{rel.name}({", ".join(args)})'

        def node_kind(node:ir.Node) -> str:
            if isinstance(node, (ir.Lookup, ir.Update, ir.Aggregate)):
                return "overload"
            return type(node).__name__.lower()

        def error_info(err:TyperError) -> str:
            if isinstance(err, TypeMismatch):
                return f'Type Mismatch|expected {to_name(err.expected)}, got {to_name(err.actual)}'
            return f'{type(err).__name__}'

        nl = "\n"
        node_strs = []
        for node in nodes:
            info = ""
            if isinstance(node, ir.Var):
                ir_type = resolved.get(node.id) or self.resolve(node)
                if not(isinstance(ir_type, ir.ScalarType) and var_is_from_literal(node)):
                    info = f'{node.name}|{type_to_str(ir_type)}'
            elif isinstance(node, ir.Literal):
                ir_type = resolved.get(node.id) or self.resolve(node)
                if not(isinstance(ir_type, ir.ScalarType)):
                    info = f'Literal|{type_to_str(ir_type)}'
            elif isinstance(node, ir.Field):
                ir_type = resolved.get(node.id) or self.resolve(node)
                info = f'{node.name}|{type_to_str(ir_type)}'
            elif isinstance(node, (ir.Lookup, ir.Update, ir.Aggregate)):
                arg_types = [self.resolve(arg) for arg in node.args]
                if node.id in self.resolved_overloads:
                    overloads = self.resolved_overloads[node.id]
                    content = nl.join([overload_str(o, arg_types) for o in overloads])
                else:
                    content = overload_str(to_relation(node), arg_types)
                info = f'{content}'
            else:
                raise NotImplementedError(f"Unknown node type: {type(node)}")

            if info:
                error_suffix = " !" if self.has_errors(node) else ""
                node_strs.append(f'{node_kind(node)}|{info}{error_suffix}')

        node_strs.sort()
        if self.errors:
            node_strs.append("---ERRORS---")
            for err in self.errors:
                node_strs.append(f'{error_info(err)} ({str(err.node).strip()})')

        return f"""{nl.join(node_strs)}"""

#--------------------------------------------------
# Analyzer
#--------------------------------------------------

class Analyzer(visitor.Visitor):
    def __init__(self):
        super().__init__()
        self.net = PropagationNetwork()
        self.context_enricher = TyperContextEnricher()

        # this is a map of literal types to a var representing that literal in
        # the graph. This allows us to collapse all literals into a single node
        # and avoids exploding node count if lots of data gets added directly
        self.shared_literal_vars:dict[ir.Type, ir.Var] = {
            t: f.var(f"__literal__{t.name}", t)
            for t in types.builtin_types
            if isinstance(t, ir.ScalarType)
        }

    #--------------------------------------------------
    # Literal types
    #--------------------------------------------------

    # given a literal type, lookup or create a var for it
    def shared_literal_var(self, node:ir.Type) -> ir.Var:
        assert isinstance(node, ir.ScalarType)
        if node not in self.shared_literal_vars:
            self.shared_literal_vars[node] = f.var(node.name, node)
        return self.shared_literal_vars[node]

    #--------------------------------------------------
    # Lookups + Aggregates
    #--------------------------------------------------

    def visit_lookup(self, node: ir.Lookup, parent: Optional[ir.Node]):
        self.visit_rel_op(node, parent)

    def visit_aggregate(self, node: ir.Aggregate, parent: ir.Node | None):
        self.visit_rel_op(node, parent)

    def visit_rel_op(self, node: ir.Lookup|ir.Aggregate, parent: Optional[ir.Node]):
        rel = to_relation(node)
        if isinstance(node, ir.Lookup) and builtins.is_eq(rel):
            return self.visit_eq(node, parent)
        if isinstance(node, ir.Lookup) and rel == builtins.cast:
            return self.visit_cast(node, parent)

        if rel.overloads:
            return self.visit_overloaded(node, parent)

        # if this is a population check, then it's fine to pass a subtype in to do the check
        # e.g. Employee(Person) is a valid way to check if a person is an employee
        is_concept_lookup = helpers.is_concept_lookup(rel)
        arg_types = set(self.net.resolve(arg) for arg, field in zip(node.args, rel.fields) if field.input)
        for arg, field in zip(node.args, rel.fields):
            field_type = field.type
            arg_type = self.net.resolve(arg)
            is_var = isinstance(arg, ir.Var)
            if not type_matches(arg_type, field_type, is_concept_lookup or is_abstract(arg_type)):
                # If the rule explicitly specifies a refined type, e.g. `Adult(v::Person)`,
                #   then `v` should be treated as `Adult` (not just `Person`).
                # In that case, check the declared population type first before checking conversions.
                if is_var and self.context_enricher.concept_population_types.get(arg.id, arg_type) == field_type:
                    continue
                # Do not complain if we can convert the arg to the field type.
                if not conversion_allowed(arg, arg_type, field_type, arg_types):
                    self.net.mismatch(node, field_type, arg_type)
                    continue
            # if we have an abstract var then this field will ultimately propagate to that
            # var's type
            elif is_var and is_abstract(arg_type) and not field.input:
                self.net.add_edge(field, arg)
                continue

        return None

    #--------------------------------------------------
    # Overloads
    #--------------------------------------------------

    def visit_overloaded(self, node: ir.Lookup|ir.Update|ir.Aggregate, parent: Optional[ir.Node]):
        incoming, outgoing, required = self.net.resolve_overload_deps(node, keep_known=True)
        for arg in incoming:
            self.net.add_edge(arg, node)
        for arg in outgoing:
            self.net.add_edge(node, arg)

    #--------------------------------------------------
    # Eq
    #--------------------------------------------------

    def visit_eq(self, node: ir.Lookup, parent: Optional[ir.Node]):
        (left, right) = node.args
        left_type = self.net.resolve(left)
        right_type = self.net.resolve(right)
        if is_abstract(left_type) and is_abstract(right_type):
            assert isinstance(left, ir.Var) and isinstance(right, ir.Var)
            # if both sides are abstract, then whatever we find out about
            # either should propagate to the other
            self.net.add_edge(left, right)
            self.net.add_edge(right, left)
        elif is_abstract(left_type):
            assert isinstance(left, ir.Var)
            if isinstance(right, ir.Var):
                self.net.add_edge(right, left)
            else:
                literal_var = self.shared_literal_var(right_type)
                self.net.add_edge(literal_var, left)
        elif is_abstract(right_type):
            assert isinstance(right, ir.Var)
            if isinstance(left, ir.Var):
                self.net.add_edge(left, right)
            else:
                literal_var = self.shared_literal_var(left_type)
                self.net.add_edge(literal_var, right)
        elif not type_matches(left_type, right_type):
            if not conversion_allowed(left, left_type, right_type) and not conversion_allowed(right, right_type, left_type):
                self.net.mismatch(node, left_type, right_type)

    #--------------------------------------------------
    # Cast
    #--------------------------------------------------

    def visit_cast(self, node: ir.Lookup, parent: Optional[ir.Node]):
        # Cast has fields: to_type (input), source (input), target (output)
        assert len(node.args) == 3, f"Expected 3 arguments for cast builtin, but got: {node.args}"
        (to_type, source, target) = node.args
        assert isinstance(to_type, ir.Type), f"Invalid target type for cast: {to_type}"
        assert isinstance(target, ir.Var), f"Invalid target variable for cast: {target}"

        if not is_abstract(to_type):
            # If target is abstract, it should get the type from to_type
            target_type = self.net.resolve(target)
            if is_abstract(target_type) and isinstance(target, ir.Var):
                self.net.add_type(target, to_type, False)

    #--------------------------------------------------
    # Update
    #--------------------------------------------------

    def visit_update(self, node: ir.Update, parent: Optional[ir.Node]):
        rel = node.relation
        # if this is a population check, then it's fine to pass a subtype in to do the population
        # e.g. Employee(Person) should be a valid way to populate a person
        allow_any_parents = helpers.is_concept_lookup(rel)
        for arg, field in zip(node.args, rel.fields):
            field_type = field.type
            arg_type = self.net.resolve(arg)
            is_var = isinstance(arg, ir.Var)

            # if the arg is abstract, but the field isn't, then we need to make
            # sure that once the arg is resolved we check that it matches the field
            # type
            if is_var and is_abstract(arg_type) and not is_abstract(field_type):
                self.net.add_type_requirement(arg, field)

            # if the field is abstract, then eventually this var will help determine
            # the field's type
            elif is_abstract(field_type):
                source = self.shared_literal_var(arg_type) if not is_var else arg
                self.net.add_edge(source, field)

            elif not type_matches(arg_type, field_type, allow_expected_parents=allow_any_parents):
                # If the rule explicitly specifies a refined type, e.g. `Adult(v::Person)`,
                #   then `v` should be treated as `Adult` (not just `Person`).
                # In that case, check the declared population type first before checking conversions.
                if is_var and self.context_enricher.concept_population_types.get(arg.id, arg_type) == field_type:
                    continue
                if not conversion_allowed(arg, arg_type, field_type):
                    self.net.mismatch(node, field_type, arg_type)

#--------------------------------------------------
# Replacer
#--------------------------------------------------

# Once we've pushed all the types through the network, we need to replace the types of
# fields and vars that we may have discovered. We also need to replace overloaded lookups
# with the choosen overloads and do any conversions that are needed.
@dataclass
class Replacer(visitor.Rewriter):
    net: PropagationNetwork = dataclasses.field(init=True)
    new_relations:OrderedSet[ir.Relation] = dataclasses.field(default_factory=OrderedSet[ir.Relation])

    def handle_model(self, model: ir.Model, parent: None):
        model = super().handle_model(model, parent)
        return model.reconstruct(
            relations=model.relations | self.new_relations
        )

    def handle_field(self, node: ir.Field, parent: ir.Node):
        if node.id in self.net.resolved_types:
            return f.field(node.name, self.net.resolved_types[node.id], node.input)
        return node

    def handle_var(self, node: ir.Var, parent: ir.Node):
        if node.id in self.net.resolved_types:
            return f.var(node.name, self.net.resolved_types[node.id])
        return node

    def handle_literal(self, node: ir.Literal, parent: ir.Node):
        # Up until now, we allow literals to have a non-builtin types. This code converts the value
        # to a builtin Python type and lifts the literal to that supertype.
        t = self.net.resolved_types[node.id] if node.id in self.net.resolved_types else node.type
        v, t = self.convert_literal_value(node.value, t)
        return node.reconstruct(value=v, type=t)

    def handle_update(self, node: ir.Update, parent: ir.Node):
        node = super().handle_update(node, parent)

        # we may do conversions, so we can end up with multiple tasks
        # in this branch and we need to track what the final args are
        tasks = []
        final_args = []
        for arg, field in zip(node.args, node.relation.fields):
            arg_type = to_type(arg)
            field_type = to_type(field)
            # the typer previously made sure that this should be valid so
            # a type mismatch means we need to convert as long as this isn't
            # a type variable
            if field_type != types.EntityTypeVar and not type_matches(arg_type, field_type):
                arg_base = to_base_primitive(arg_type)
                field_base = to_base_primitive(field_type)
                if arg_base is not None and field_base is not None and conversion_allowed(arg, arg_base, field_base):
                    new_arg = self.convert(arg, arg_base, field_base, tasks)
                    final_args.append(new_arg)
                else:
                    final_args.append(arg)
            else:
                final_args.append(arg)
        tasks.append(node.reconstruct(args=tuple(final_args)))

        if len(tasks) == 1:
            return tasks[0]
        else:
            return f.logical(tasks, [])

    def handle_aggregate(self, node: ir.Aggregate, parent: ir.Node):
        node = super().handle_aggregate(node, parent)
        lookup_args = list(node.args)
        hoists = helpers.get_agg_outputs(node)

        if len(node.aggregation.fields) == len(lookup_args):
            tasks = []
            for i in range(len(lookup_args)):
                if not conversion_allowed(lookup_args[i], to_type(lookup_args[i]), node.aggregation.fields[i].type):
                    self.net.mismatch(node, node.aggregation.fields[i].type, to_type(lookup_args[i]))
                elif not type_matches(to_type(lookup_args[i]), node.aggregation.fields[i].type):
                    lookup_args[i] = self.convert(lookup_args[i], to_type(lookup_args[i]), node.aggregation.fields[i].type, tasks)
            tasks.append(node.reconstruct(args=tuple(lookup_args)))
            if len(tasks) == 1:
                return tasks[0]
            else:
                return f.logical(tasks, hoists)

        return node

    def handle_lookup(self, node: ir.Lookup, parent: ir.Node):
        node = super().handle_lookup(node, parent)
        lookup_args = node.args

        # we can resolve to multiple overloads, that we'd need to union
        # so we'll need to hoist all the vars being output to make sure
        # everything joins correctly
        hoists = helpers.get_outputs(node)
        branches:list[ir.Logical] = []

        # Special case this solver primitive because the type system cannot handle what it
        # needs, which is an input field types as List[Union[Int64,Float64,String,Hash]].
        # So this code makes sure that any in128 that flows into the second argument, which
        # is represented as a tuple, is converted into an int64, which is lossy but should
        # work for most cases
        if node.relation == builtins.rel_primitive_solverlib_fo_appl and len(node.relation.fields) == 3 and len(lookup_args) == 3:
            tasks = []
            # Make lookup_args mutable.
            lookup_args = list(lookup_args)

            # Convert the first and third arguments to the field types.
            for i in [0, 2]:
                if not conversion_allowed(lookup_args[i], to_type(lookup_args[i]), node.relation.fields[i].type):
                    self.net.mismatch(node, node.relation.fields[i].type, to_type(lookup_args[i]))
                elif not type_matches(to_type(lookup_args[i]), node.relation.fields[i].type):
                    lookup_args[i] = self.convert(lookup_args[i], to_type(lookup_args[i]), node.relation.fields[i].type, tasks)

            varargs = lookup_args[1]
            final_varargs = []
            assert isinstance(varargs, tuple)
            for arg in varargs:
                t = to_type(arg)
                if t == types.Int128:
                    final_varargs.append(self.convert(arg, t, types.Int64, tasks))
                else:
                    final_varargs.append(arg)
            lookup_args[1] = tuple(final_varargs)
            tasks.append(node.reconstruct(args=tuple(lookup_args)))
            branches.append(f.logical(tasks, hoists))

        elif builtins.is_eq(node.relation):
            tasks = []

            # We need to handle eq specially because its arguments can be converted symmetrically
            (left, right) = lookup_args
            left_type = to_type(left)
            right_type = to_type(right)
            mismatch = False
            if not type_matches(left_type, right_type):
                if conversion_allowed(left, left_type, right_type):
                    new_left = self.convert(left, left_type, right_type, tasks)
                    final_args = [new_left, right]
                elif conversion_allowed(right, right_type, left_type):
                    new_right = self.convert(right, right_type, left_type, tasks)
                    final_args = [left, new_right]
                else:
                    self.net.mismatch(node, left_type, right_type)
                    final_args = [left, right]
                    mismatch = True
            else:
                final_args = [left, right]

            if not mismatch:
                min_cost = float('inf')
                arg_types = set(to_type(arg) for arg in final_args)
                resolved = []
                for o in node.relation.overloads:
                    total = 0
                    for arg, field in zip(final_args, o.fields):
                        arg_type = to_type(arg)
                        field_type = to_type(field)
                        total += conversion_cost(arg, arg_type, field_type, arg_types)
                    if total <= min_cost:
                        if total < min_cost:
                            resolved.clear()
                            min_cost = total
                        resolved.append(o)

                if len(resolved) == 1:
                    tasks.append(f.lookup(resolved[0], final_args))
                else:
                    # If we cannot resolve the overload, just leave it.
                    self.net.unresolved_overload(node)
                    tasks.append(f.lookup(node.relation, final_args))
            else:
                # If there's a mismatch, just leave the original relation.
                tasks.append(f.lookup(node.relation, final_args))

            branches.append(f.logical(tasks, hoists))
        elif node.relation == builtins.cast:
            assert len(node.args) == 3, f"Invalid number of arguments for cast: {node.args}"
            (tgt_type, src, tgt) = node.args
            assert isinstance(tgt_type, ir.Type), f"Invalid target type for cast: {tgt_type}"
            src_type = to_type(src)

            # if we are casting a literal into a var, we can just set the var to a new literal
            # with the same value but with the same type as the variable
            if isinstance(src, ir.Literal) and literal_conversion_allowed(src_type, tgt_type):
                annos = node.annotations
                return f.lookup(builtins.eq, (ir.Literal(tgt_type, src.value), tgt), annos=(list(annos) + [builtins.from_cast_annotation]))

            src_base = to_base_primitive(src_type)
            tgt_base = to_base_primitive(tgt_type)

            if src_base and tgt_base and src_base == tgt_base:
                annos = node.annotations
                return f.lookup(builtins.eq, (src, tgt), annos=(list(annos) + [builtins.from_cast_annotation]))

            if conversion_allowed(node.args[0], src_type, tgt_type):
                if tgt_base:
                    return node.reconstruct(args=(tgt_base, src, tgt))
                else:
                    return node
            else:
                # TODO: we are simply ignoring this, assuming there's some way to cast it
                # self.net.mismatch(node, src_type, tgt_type)
                return node
        else:
            if node.id in self.net.resolved_overloads:
                resolved = self.net.resolved_overloads[node.id]
            else:
                resolved = [node.relation]
            decimal_type = self.net.resolved_overload_decimal.get(node.id)

            for overload in resolved:
                # we may do conversions, so we can end up with multiple tasks
                # in this branch and we need to track what the final args are
                tasks = []
                final_args = []

                for arg, field in zip(lookup_args, overload.fields):
                    arg_type = to_type(arg)
                    field_type = to_type(field)
                    # the typer previously made sure that this should be valid so
                    # a type mismatch means we need to convert as long as this isn't
                    # a type variable
                    if field_type == types.GenericDecimal or (field_type != types.EntityTypeVar and not arg_type == field_type):
                        arg_base = to_base_primitive(arg_type)
                        field_base = to_base_primitive(field_type)
                        if field_base == types.GenericDecimal:
                            field_base = decimal_type
                        if arg_base is not None and field_base is not None and conversion_allowed(arg, arg_base, field_base):
                            new_arg = self.convert(arg, arg_base, field_base, tasks)
                            final_args.append(new_arg)
                        else:
                            final_args.append(arg)
                    else:
                        final_args.append(arg)
                tasks.append(f.lookup(overload, final_args))
                branches.append(f.logical(tasks, hoists))
        # unwrap if we don't actually need a union
        if len(branches) == 1:
            if len(branches[0].body) == 1:
                return branches[0].body[0]
            else:
                return branches[0]
        else:
            return f.union(branches, hoists)

    def convert(self, arg:ir.Value, actual:ir.Type, expected:ir.Type, tasks:list[ir.Task]) -> ir.Value:
        if actual == expected:
            return arg
        value = None
        if isinstance(arg, ir.Literal):
            value = arg.value
        elif isinstance(arg, (int, float, PyDecimal)):
            value = arg
        if value is not None:
            if isinstance(value, int) and expected in (types.Int64, types.Int128):
                if expected == types.Int64:
                    check_int64(value)
                return f.literal(value, expected)
            if isinstance(value, (int, float, PyDecimal)) and expected == types.Float:
                return f.literal(float(value), expected)
            if isinstance(value, (int, float, PyDecimal)) and types.is_decimal(expected):
                # Converting str(value) rather than value avoids precision loss with Float->Decimal conversion.
                return f.literal(PyDecimal(str(value)), expected)

        name = helpers.sanitize(arg.name + "_" + to_name(expected) if isinstance(arg, ir.Var) else f"v{to_name(expected)}")
        expected_base = to_base_primitive(expected) or expected
        new_arg = f.var(name, expected_base)
        self.new_relations.add(builtins.cast)
        tasks.append(f.lookup(builtins.cast, (expected_base, arg, new_arg)))
        self.new_relations.add(builtins.cast)
        return new_arg

    @staticmethod
    def convert_literal_value(value, t: ir.Type):
        if types.is_subtype(t, types.String):
            return value, types.String
        if isinstance(value, str) and types.is_subtype(t, types.DateTime):
            return datetime.datetime.fromisoformat(value), types.DateTime
        if isinstance(value, str) and types.is_subtype(t, types.Date):
            return datetime.date.fromisoformat(value), types.Date
        if isinstance(value, int) and types.is_subtype(t, types.Int64):
            check_int64(value)
            return value, types.Int64
        if isinstance(value, int) and types.is_subtype(t, types.Int128):
            return value, types.Int128
        if isinstance(value, (int, float, PyDecimal)) and types.is_subtype(t, types.Float):
            return float(value), types.Float
        if isinstance(value, (int, float, PyDecimal)) and types.is_decimal_subtype(t):
            return PyDecimal(str(value)), types.decimal_supertype(t)
        return value, t

#--------------------------------------------------
# Typer pass
#--------------------------------------------------

class InferTypes(compiler.Pass):
    def __init__(self):
        super().__init__()
        self.historical = {}

    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        w = Analyzer()
        w.net.load_types(self.historical)

        # collect some type info before analyzing it
        with debugging.span("type.collect"):
            model.root.accept(w.context_enricher, model)

        # build the propagation network
        with debugging.span("type.analyze"):
            model.root.accept(w, model)

        # propagate the types through the network
        with debugging.span("type.propagate") as end_span:
            field_types = w.net.propagate()
            self.historical.update(field_types)
            end_span["type_graph"] = w.net.to_mermaid()
            end_span["type_report"] = w.net.to_fish()

        # replace the fields in the model with the new types
        with debugging.span("type.replace"):
            final = Replacer(w.net).walk(model)

        for err in w.net.errors:
            rich.print(str(err), file=sys.stderr)

        return final



import warnings

from v0.relationalai.semantics.metamodel import (types, util, ir, factory, helpers, builtins, compiler, executor, dependency,
                                              Node, Model, Capability, Engine, Type, ScalarType, ListType, UnionType,
                                              Field, Relation, Task, Logical, Union, Sequence, Match, Until, Wait, Not,
                                              Exists, ForAll, Loop, Break, Var, Literal, Value, Annotation, Effect,
                                              Update, Lookup, Output, Construct, Aggregate)

__all__ = [
    "types",
    "util",
    "ir",
    "factory",
    "helpers",
    "builtins",
    "compiler",
    "executor",
    "dependency",
    "Node",
    "Model",
    "Capability",
    "Engine",
    "Type",
    "ScalarType",
    "ListType",
    "UnionType",
    "Field",
    "Relation",
    "Task",
    "Logical",
    "Union",
    "Sequence",
    "Match",
    "Until",
    "Wait",
    "Not",
    "Exists",
    "ForAll",
    "Loop",
    "Break",
    "Var",
    "Literal",
    "Value",
    "Annotation",
    "Effect",
    "Update",
    "Lookup",
    "Output",
    "Construct",
    "Aggregate",
]

warnings.warn(
    "relationalai.early_access.metamodel.* is deprecated. "
    "Please migrate to relationalai.semantics.metamodel.*",
    DeprecationWarning,
    stacklevel=2,
)



import logging
import os
import uuid
import sys
import datetime
from abc import abstractmethod, ABC

from v0.relationalai.semantics.tests.logging import Capturer
from v0.relationalai.semantics.tests.utils import reset_state
from v0.relationalai.semantics.internal import internal
from v0.relationalai.clients.result_helpers import sort_data_frame_result
from v0.relationalai.clients.util import IdentityParser
from v0.relationalai.clients.resources.snowflake import Provider as SFProvider
from v0.relationalai import Provider
from typing import cast, Dict, Union
from pathlib import Path

class AbstractSnapshotTest(ABC):

    provider:Provider = cast(SFProvider, Provider()) # type: ignore

    def run_snapshot_test(self, snapshot, script_path, db_schema=None, use_sql=False, use_lqp=True, use_rel=False,
                          use_direct_access=False, e2e=False, use_csv=True, e2e_only=False):
        # Resolve use_lqp
        use_lqp = use_lqp and (not use_rel) # use_rel overrides because use_lqp is default.

        """Run a snapshot test.

        This method assumes content is already in the correct string format.
        Subclasses should handle any necessary compilation or conversion.
        """
        logger = logging.getLogger("pyrellogger")
        handler = Capturer()
        logger.addHandler(handler)

        # reset state across test runs
        reset_state()

        # Track which modules are already loaded
        before_modules = set(sys.modules.keys())
        unique_name = f"md{str(uuid.uuid4())[-12:]}"
        export_fqn = None
        dry_run = not e2e
        try:
            overrides = {
                'dry_run': dry_run,
                'model_suffix': "" if not e2e else f"_{unique_name}",
                'use_sql': use_sql,
                'reasoner.rule.use_lqp': use_lqp,
                'keep_model': False,
                # fix the current time to keep snapshots stable
                'datetime_now': datetime.datetime.fromisoformat("2025-12-01T12:00:00+00:00"),
            }
            if use_direct_access:
                # for direct access we can not use user/password authentication
                # so we override the config to use key pair authentication
                override_config: Dict[str, Union[str, bool]] = {
                    'use_direct_access': True,
                }
                if os.getenv("AUTHENTICATOR"):
                    override_config["authenticator"] = os.getenv("AUTHENTICATOR", "")
                if os.getenv("PRIVATE_KEY_FILE"):
                    override_config["private_key_file"] = os.getenv("PRIVATE_KEY_FILE", "")
                if os.getenv("SF_TEST_ACCOUNT_KEY_PASSPHRASE"):
                    override_config["private_key_file_pwd"] = os.getenv("SF_TEST_ACCOUNT_KEY_PASSPHRASE", "")
                overrides['config'] = override_config

            with internal.with_overrides(**overrides):
                with open(script_path, "r") as f:
                    # Compile with the correct filename
                    code = compile(f.read(), script_path, "exec")
                    # Execute the compiled code
                    if db_schema:
                        export_fqn = f"{db_schema}.{unique_name}"
                    exec_globals = {"EXPORT_TABLE": export_fqn}
                    exec(code, exec_globals)
        finally:
            # Remove any modules that were imported during exec()
            after_modules = set(sys.modules.keys())
            new_modules = after_modules - before_modules
            for mod in new_modules:
                sys.modules.pop(mod, None)
            # Remove handler
            logger.removeHandler(handler)
            # check if script created an export table
            export_exists = not dry_run and self.is_export_exists(export_fqn)
            # match snapshots
            self.assert_match_snapshots(script_path, snapshot, handler, export_exists, export_fqn, use_sql=use_sql,
                                        use_lqp=use_lqp, e2e=e2e, use_csv=use_csv, e2e_only=e2e_only)
            # cleanup resources created during a test run
            self.cleanup(export_exists, export_fqn)

    def assert_match_snapshots(self, script_path, snapshot, handler, export_exists=False, export_fqn=None, use_sql=False,
                               use_lqp=True, e2e=False, use_csv=True, e2e_only=False):
        if not e2e_only:
            self.assert_match_internal_results_snapshots(snapshot, handler, use_sql, use_lqp)
        if e2e:
            self.assert_match_results_snapshots(script_path, snapshot, handler)
        if export_fqn and export_exists:
            result_df = sort_data_frame_result(self.provider.sql(f"SELECT * FROM {export_fqn}", format="pandas"))
            if use_csv:
                snapshot.assert_match(result_df.to_csv(index=False).strip('"').strip('\n"'), "export_result.csv")
            else:
                snapshot.assert_match(result_df.to_string(), "export_result.df")

    def assert_match_internal_results_snapshots(self, snapshot, handler, use_sql, use_lqp):
        # default: do nothing
        return

    @abstractmethod
    def assert_match_results_snapshots(self, script_path, snapshot, handler):
        pass

    def cleanup(self, export_exists=False, export_fqn=None):
        if export_fqn and export_exists :
            self.provider.sql(f"DROP TABLE IF EXISTS {export_fqn}")

    def is_export_exists(self, export_fqn=None):
        if export_fqn is None:
            return False
        database, schema, table, _ = IdentityParser(export_fqn, require_all_parts=True).to_list()
        return self.provider.sql(f"""
            SELECT EXISTS(
                SELECT 1
                FROM {database}.INFORMATION_SCHEMA.TABLES
                WHERE TABLE_SCHEMA = '{schema}'
                AND TABLE_NAME = '{table}'
            ) AS TABLE_EXISTS;
        """)[0]["TABLE_EXISTS"]

    @staticmethod
    def parent_tests(file: str):
        return AbstractSnapshotTest.discover_test_scripts(Path(file).parent)

    @staticmethod
    def discover_test_scripts(path: Path):
        scripts_dir = path / "tests"
        return sorted(scripts_dir.glob("**/*.py"))

    @staticmethod
    def get_id(p):
        p_name = p.parent.name
        return p.stem if p_name == "tests" else f"{p_name}_{p.stem}"



from __future__ import annotations
from typing import Optional, Sequence
from v0.relationalai.semantics.internal import internal
from v0.relationalai.semantics.metamodel.ir import (
    Node, Require, Logical, Var, Relation, Lookup, ScalarType
)
from v0.relationalai.semantics.metamodel import builtins


"""
Helper functions for converting `Require` nodes with unique constraints to functional
dependencies. The main functionalities provided are:
    1. Check whether a `Require` node is a valid unique constraint representation
    2. Represent the uniqueness constraint as a functional dependency
    3. Check if the functional dependency is structural i.e., can be represented with
         `@function(k)` annotation on a single relation.

=========================== Structure of unique constraints ================================
A `Require` node represents a _unique constraint_ if it meets the following criteria:
 * the `Require` node's `domain` is an empty `Logical` node
 * the `Require` node's `checks` has a single `Check` node
 * the single `Check` node has `Logical` task that is a list of `Lookup` tasks
 * precisely one `Lookup` task in the `Check` uses the `unique` builtin relation name
 * the `unique` lookup has precisely one argument, which is a `TupleArg` or a `tuple`
   containing at least one `Var`
 * all `Lookup` nodes use variables only (no constants)
 * the variables used in the `unique` lookup are a subset of the variables used in other
   lookups
============================================================================================

We use the following unique constraint as the running example.

```
Require
  domain
    Logical
  checks:
    Check
      check:
        Logical
          Person(person::Person)
          first_name(person::Person, firstname::String)
          last_name(person::Person, lastname::String)
          unique((firstname::String, lastname::String))
      error:
        ...
```

=========================== Semantics of unique constraints ================================
A unique constraint states that the columns declared in the `unique` predicate must be
unique in the result of the conjunctive query consisting of all remaining predicates.
============================================================================================

In the running example, the conjunctive query computes a table with 3 columns, the person id
`person::Person`, the first name `firstname::String`, and the last name `lastname::String`.
The uniqueness predicate `unique((firstname::String, lastname::String))` states that no person
can have more than a single combination of first and last name.

The unique constraint in the running example above corresponds to the following functional
dependency.

```
Person(x) ∧ first_name(x, y) ∧ last_name(x, z): {y, z} -> {x}
```

------------------------------ Redundant Type Atoms ----------------------------------------
At the time of writing, PyRel does not yet remove redundant unary atoms. For instance, in
the running example, the atom `Person(person::Person)` is redundant because the type of the
`person` variable is specified in the other two atoms `first_name` and `last_name`.
Consequently, we identify redundant atoms and remove them from the definition of the
corresponding functional dependency.

Formally, a _guard_ atom is any `Lookup` node whose relation name is not `unique`. Now, a
unary guard atom `T(x::T)` is _redundant_ if the uniqueness constraint has a non-unary guard
atom `R(...,x::T,...)`.

================================ Normalized FDs ============================================
Now, the _(normalized)_ functional dependency_ corresponding to a unique constraint is an
object of the form `φ: X → Y`, where :
 1. `φ` is the set of all non-redundant guard atoms.
 2. `X` is the set of variables used in the `unique` atom
 3. `Y` is the set of all other variables used in the constraint
============================================================================================

The normalized functional dependency corresponding to the unique constraints from the running
example is :
```
first_name(person::Person, firstname::String) ∧ last_name(person::Person, lastname::String): {firstname:String, lastname:String} -> {person:Person}
```
Note that the unary atom `Person(person::Person)` is redundant and thus omitted from the
decomposition.

Some simple functional dependencies can, however, be expressed simply with `@function(k)`
attribute of a single relation.  Specifically, a functional dependency `φ: X → Y` is
_structural_ if φ consists of a single atom `R(x1,...,xm,y1,...,yk)` and `X = {x1,...,xm}`.
"""

#
# Checks that an input `Require` node is a valid unique constraint. Returns `None` if not.
# If yes, we return the decomposition of the unique constraint as a tuple
# `(all_vars, unique_vars, guard)`, where
# - `all_vars` is the list of all variables used in the constraint
# - `unique_vars` is the list of variables used in the `unique` atom
# - `guard` is the list of all other `Lookup` atoms
#
def _split_unique_require_node(node: Require) -> Optional[tuple[list[Var], list[Var], list[Lookup]]]:
    if not isinstance(node.domain, Logical):
        return None
    if len(node.domain.body) != 0:
        return None
    if len(node.checks) != 1:
        return None
    check = node.checks[0]
    if not isinstance(check.check, Logical):
        return None

    unique_atom: Optional[Lookup] = None
    guard: list[Lookup] = []
    for task in check.check.body:
        if not isinstance(task, Lookup):
            return None
        if task.relation.name == builtins.unique.name:
            if unique_atom is not None:
                return None
            unique_atom = task
        else:
            guard.append(task)

    if unique_atom is None:
        return None

    # collect variables
    all_vars: list[Var] = []
    for lookup in guard:
        for arg in lookup.args:
            if not isinstance(arg, Var):
                return None
            if arg in all_vars:
                continue
            all_vars.append(arg)

    unique_vars: list[Var] = []
    if len(unique_atom.args) != 1:
        return None
    if not isinstance(unique_atom.args[0], (internal.TupleArg, tuple)):
        return None
    if len(unique_atom.args[0]) == 0:
        return None
    for arg in unique_atom.args[0]:
        if not isinstance(arg, Var):
            return None
        if arg in unique_vars:
            return None
        unique_vars.append(arg)

    # check that unique vars are a subset of other vars
    if not set(unique_vars).issubset(set(all_vars)):
        return None

    return list(all_vars), list(unique_vars), guard


def is_valid_unique_constraint(node: Require) -> bool:
    """
    Checks whether the input `Require` node is a valid unique constraint. See description at
    the top of the file for details.
    """
    return _split_unique_require_node(node) is not None

#
# A unary guard atom `T(x::T)` is redundant if the constraint contains a non-unary atom
# `R(...,x::T,...)`.  We discard all redundant guard atoms in the constructed fd.
#
def normalized_fd(node: Require) -> Optional[FunctionalDependency]:
    """
    If the input `Require` node is a uniqueness constraint, constructs its reduced
    functional dependency `φ: X -> Y`, where `φ` contains all non-redundant guard atoms,
    `X` are the variables used in the `unique` atom, and `Y` are the remaining variables.
    Returns `None` if the input node is not a valid uniqueness constraint.
    """
    parts = _split_unique_require_node(node)
    if parts is None:
        return None
    all_vars, unique_vars, guard_atoms = parts

    # remove redundant lookups
    redundant_guard_atoms: list[Lookup] = []
    for atom in guard_atoms:
        # the atom is unary A(x::T)
        if len(atom.args) != 1:
            continue
        var = atom.args[0]
        assert isinstance(var, Var)
        # T is a scalar type (which includes entity types)
        var_type = var.type
        if not isinstance(var_type, ScalarType):
            continue
        # the atom is a entity typing T(x::T) i.e., T = A (and hence not a Boolean property)
        var_type_name = var_type.name
        rel_name = atom.relation.name
        if rel_name != var_type_name:
            continue
        # Found an atom of the form T(x::T)
        # check if there is another atom R(...,x::T,...)
        for typed_atom in guard_atoms:
            if len(typed_atom.args) == 1:
                continue
            if var in typed_atom.args:
                redundant_guard_atoms.append(atom)
                break

    guard = [atom for atom in guard_atoms if atom not in redundant_guard_atoms]
    keys = unique_vars
    values = [v for v in all_vars if v not in keys]

    return FunctionalDependency(guard, keys, values)

class FunctionalDependency:
    """
    Represents a functional dependency of the form `φ: X -> Y`, where
     - `φ` is a set of `Lookup` atoms
     - `X` and `Y` are disjoint and covering sets of variables used in `φ`
    """
    def __init__(self, guard: Sequence[Lookup], keys: Sequence[Var], values: Sequence[Var]):
        self.guard = tuple(guard)
        self.keys = tuple(keys)
        self.values = tuple(values)
        assert set(self.keys).isdisjoint(set(self.values)), "Keys and values must be disjoint"

        # for structural fd check
        self._is_structural:bool = False
        self._structural_relation:Optional[Relation] = None
        self._structural_rank:Optional[int] = None

        self._determine_is_structural()

    # A functional dependency `φ: X → Y` is _k-functional_ if `φ` consists of a single atom
    # `R(x1,...,xm,y1,...,yk)` and `X = {x1,...,xm}`. Not all functional dependencies are
    # k-functional. For instance, `R(x, y, z): {y, z} → {x}` cannot be expressed with
    # `@function`. neither can `R(x, y) ∧ P(x, z) : {x} → {y, z}`.
    def _determine_is_structural(self):
        if len(self.guard) != 1:
            self._is_structural = False
            return
        atom = next(iter(self.guard))
        atom_vars = atom.args
        if len(atom_vars) <= len(self.keys): # @function(0) provides no information
            self._is_structural = False
            return
        prefix_vars = atom_vars[:len(self.keys)]
        if set(prefix_vars) != set(self.keys):
            self._is_structural = False
            return
        self._is_structural = True
        self._structural_relation = atom.relation
        self._structural_rank = len(atom_vars) - len(self.keys)

    @property
    def is_structural(self) -> bool:
        """
        Whether the functional dependency is functional, i.e., can be represented
        with `@function(k)` annotation on a single relation.
        """
        return self._is_structural

    @property
    def structural_relation(self) -> Relation:
        """
        The structural relation of a functional dependency. Raises ValueError if the functional
        dependency is not structural.
        """
        if not self._is_structural:
            raise ValueError("Functional dependency is not structural")
        assert self._structural_relation is not None
        return self._structural_relation

    @property
    def structural_rank(self) -> int:
        """
        The structural rank k of k-structural fd. Raises ValueError if the structural
        dependency is not k-structural.
        """
        if not self._is_structural:
            raise ValueError("Functional dependency is not structural")
        assert self._structural_rank is not None
        return self._structural_rank

    def __str__(self) -> str:
        guard_str = " ∧ ".join([str(atom) for atom in self.guard]).strip()
        keys_str = ", ".join([str(var) for var in self.keys]).strip()
        values_str = ", ".join([str(var) for var in self.values]).strip()
        return f"{guard_str}: {{{keys_str}}} -> {{{values_str}}}"

def contains_only_declarable_constraints(node: Node) -> bool:
    """
    Checks whether the input `Logical` node contains only `Require` nodes annotated with
    `declare_constraint`.
    """
    if not isinstance(node, Logical):
        return False
    if len(node.body) == 0:
        return False
    for task in node.body:
        if not isinstance(task, Require):
            return False
        if not is_declarable_constraint(task):
            return False
    return True

def is_declarable_constraint(node: Require) -> bool:
    """
    Checks whether the input `Require` node is annotated with `declare_constraint`.
    """
    return builtins.declare_constraint_annotation in node.annotations



from __future__ import annotations
import ast
from dataclasses import dataclass, field
from functools import lru_cache
import os
import sys
import textwrap
import dis
from bisect import bisect_right
from types import CodeType

cwd = os.getcwd()

#------------------------------------------------------
# Path utilities
#------------------------------------------------------

# relpath is a shockingly slow operation, so caching it
@lru_cache(maxsize=128)
def relative_path(path: str) -> str:
    try:
        return os.path.relpath(path, cwd)
    except ValueError:
        return path

#------------------------------------------------------
# Frame utilities
#------------------------------------------------------

_line_map_cache = {}

def instruction_line(code, lasti):
    pairs = _line_map_cache.get(code)
    if pairs is None:
        pairs = list(dis.findlinestarts(code))  # (addr, line_number)
        _line_map_cache[code] = pairs

    idx = bisect_right(pairs, (lasti, float("inf"))) - 1
    if idx < 0:
        return code.co_firstlineno
    return pairs[idx][1]

def first_non_relationalai_frame():
    frame = sys._getframe(2)  # skip this function and the previous one
    while frame and frame.f_back:
        mod_name = frame.f_globals.get("__name__", "")
        if mod_name.startswith("relationalai") or mod_name.startswith("v0.relationalai"):
            frame = frame.f_back
            continue
        return frame
    return frame

#------------------------------------------------------
# Source utilities
#------------------------------------------------------

@lru_cache(maxsize=128)
def get_source_code(filename: str) -> str|None:
    """Get the source code of a file."""
    try:
        with open(filename, "r") as f:
            return f.read()
    except Exception:
        return None

parses = {}
def find_root_expression(source_code: str, target_line: int, relative_filename: str):
    # Parse the source code into an AST
    if relative_filename not in parses:
        parses[relative_filename] = ast.parse(source_code)
    tree = parses[relative_filename]

    class ExpressionChainFinder(ast.NodeVisitor):
        def __init__(self, target_line):
            self.target_line = target_line
            self.candidates = []

        def visit(self, node):
            # Consider all expression nodes, including Call nodes
            is_expr_node = isinstance(node, (ast.Expr, ast.Call, ast.Attribute, ast.BinOp,
                                          ast.UnaryOp, ast.IfExp, ast.Compare, ast.Subscript,
                                          ast.Dict, ast.List, ast.Tuple, ast.Set))

            if hasattr(node, "lineno") and is_expr_node:
                end_lineno = getattr(node, "end_lineno", node.lineno)

                # Check if this node contains our target line
                if node.lineno <= self.target_line <= end_lineno:
                    # Store the node with its line span information
                    self.candidates.append((node, node.lineno, end_lineno))

            # Continue visiting children
            for child in ast.iter_child_nodes(node):
                self.visit(child)

    finder = ExpressionChainFinder(target_line)
    finder.visit(tree)

    if finder.candidates:
        # Sort candidates by their starting line (ascending) and then by span size (descending)
        # This prioritizes expressions that start earlier and cover more lines
        sorted_candidates = sorted(finder.candidates,
                                  key=lambda x: (x[1], -(x[2] - x[1])))

        # Find the outermost expression that contains our target line
        # This will capture the entire multi-line function call
        outermost_candidates = []
        for node, start_line, end_line in sorted_candidates:
            # Check if this is a new candidate or extends an existing one
            if not outermost_candidates or start_line < outermost_candidates[0][1]:
                outermost_candidates = [(node, start_line, end_line)]

            # If we have a node that starts at the same line but ends later, prefer it
            elif start_line == outermost_candidates[0][1] and end_line > outermost_candidates[0][2]:
                outermost_candidates = [(node, start_line, end_line)]

        if outermost_candidates:
            node, start_line, end_line = outermost_candidates[0]

            # Extract the lines from the source code
            block_lines = source_code.splitlines()[start_line - 1:end_line]
            block_code = "\n".join(block_lines)
            return SourcePos(relative_filename, _line=start_line, _source=textwrap.dedent(block_code))

    # If no nodes were found, return the single line
    lines = source_code.splitlines()
    if target_line > len(lines):
        return SourcePos(relative_filename, _line=target_line)
    return SourcePos(relative_filename, _line=target_line, _source=lines[target_line - 1])

#------------------------------------------------------
# SourcePos
#------------------------------------------------------

counts = 0

@dataclass(slots=True)
class SourcePos:
    file: str = field(default="Unknown")
    _block:SourcePos|None = None
    _lasti: int|None = field(default=0)
    _pycode: CodeType|None = field(default=None)
    _source: str|None = field(default=None)
    _line: int|None = field(default=None)

    @classmethod
    def new(cls):
        caller_frame = first_non_relationalai_frame()
        if not caller_frame:
            return cls()

        filename = caller_frame.f_code.co_filename
        lasti = caller_frame.f_lasti
        pycode = caller_frame.f_code

        return cls(filename, _lasti=lasti, _pycode=pycode)

    @property
    def line(self):
        if self._line is None and self._pycode:
            self._line = instruction_line(self._pycode, self._lasti)
        return self._line or 0

    @property
    def source(self):
        if self._source is None and self.file:
            self._source = get_source_code(self.file)
        return self._source or ""

    @property
    def block(self):
        if not self.source or not self.line:
            return SourcePos()
        if not self._block:
            self._block = find_root_expression(self.source, self.line, self.file)
        return self._block

    def transform(self, transformer:ast.NodeTransformer) -> SourcePos|None:
        if not self.block or not self.block.source:
            return

        new_ast = transformer.visit(ast.parse(self.block.source))
        new_source = ast.unparse(new_ast)
        new = SourcePos(self.file, _line=self.line, _source=new_source)
        return new



# debug_snapshot.py
from __future__ import annotations
import os, json, time, tempfile
from typing import List

def _tail_lines(path: str, max_lines: int) -> List[str]:
    """Tail last N lines efficiently; fine for small/medium files."""
    if not os.path.exists(path): return []
    # simple approach: read and slice; optimize later if needed
    with open(path, "r", encoding="utf-8", errors="ignore") as f:
        return f.readlines()[-max_lines:]

def write_snapshot(spans_jsonl: str, out_html: str, last: int = 5000):
    lines = _tail_lines(spans_jsonl, last)
    spans = [json.loads(x) for x in lines if x.strip()]
    html = _HTML_TEMPLATE.replace("/*__EMBED__*/", json.dumps(spans, separators=(",", ":")))
    os.makedirs(os.path.dirname(out_html) or ".", exist_ok=True)
    fd, tmp = tempfile.mkstemp(dir=os.path.dirname(out_html) or ".", prefix=".part-", suffix=".html")
    with os.fdopen(fd, "w", encoding="utf-8") as f:
        f.write(html); f.flush(); os.fsync(f.fileno())
    os.replace(tmp, out_html)

def start_background_snapshots(spans_jsonl: str, out_html: str, interval_ms: int = 1000, last: int = 5000):
    import threading, atexit
    stop = threading.Event()

    def loop():
        next_t = time.monotonic()
        while not stop.is_set():
            try:
                write_snapshot(spans_jsonl, out_html, last=last)
            except Exception:
                pass
            next_t += interval_ms / 1000.0
            delay = max(0.0, next_t - time.monotonic())
            stop.wait(delay)

    th = threading.Thread(target=loop, name="debug-snapshot", daemon=True)
    th.start()
    atexit.register(stop.set)
    return stop

_HTML_TEMPLATE = """<!doctype html>
<meta charset="utf-8">
<title>Rel Debugger</title>
<meta http-equiv="refresh" content="1">
<style>
  :root { --ok:#22c55e; --err:#ef4444; --run:#a3a3a3; --bg:#0b0c0e; --fg:#e5e7eb; --mut:#9ca3af; }
  html, body { height:100%; background:var(--bg); color:var(--fg); font:14px ui-sans-serif, system-ui, -apple-system, Segoe UI, Roboto; margin:0; }
  header { padding:10px 14px; border-bottom:1px solid #1f2937; display:flex; gap:12px; align-items:center; }
  .chip { padding:2px 8px; border-radius:999px; background:#111827; color:var(--mut); border:1px solid #1f2937; }
  .wrap { display:grid; grid-template-columns: 380px 1fr; height:calc(100% - 48px); }
  aside { overflow:auto; border-right:1px solid #1f2937; }
  main { overflow:auto; position:relative; }
  ul { list-style:none; margin:0; padding-left:12px; }
  li { padding:4px 8px; border-left:2px solid transparent; }
  li.ok { border-color: var(--ok); }
  li.err { border-color: var(--err); }
  li.run { border-color: var(--run); }
  .name { font-weight:600; }
  .meta { color:var(--mut); font-size:12px; }
  .tl { position:relative; margin:10px; height:24px; background:#111827; border:1px solid #1f2937; border-radius:6px; }
  .bar { position:absolute; top:0; bottom:0; border-radius:6px; }
  .ok .bar { background:linear-gradient(90deg, #14532d, var(--ok)); }
  .err .bar { background:linear-gradient(90deg, #7f1d1d, var(--err)); }
  .run .bar { background:linear-gradient(90deg, #374151, var(--run)); }
  details > summary { cursor:pointer; }
  code { background:#111827; border:1px solid #1f2937; padding:1px 4px; border-radius:4px; }
</style>
<header>
  <div class="chip" id="c-traces">0 traces</div>
  <div class="chip" id="c-spans">0 spans</div>
  <div class="chip" id="c-open">0 running</div>
  <div class="chip" id="c-warn">0 warnings</div>
</header>
<div class="wrap">
  <aside id="tree"></aside>
  <main id="timelines"></main>
</div>

<script>
const STATE_KEY = `rel-debugger:${location.pathname}`;

function loadState(){
  try {
    const raw = localStorage.getItem(STATE_KEY);
    return raw ? JSON.parse(raw) : {};
  } catch (_) {
    return {};
  }
}

const state = loadState();
if (!state.scroll || typeof state.scroll !== 'object') state.scroll = {};
if (!Array.isArray(state.detailsOpen)) state.detailsOpen = [];
const detailsOpen = new Set(state.detailsOpen);
let saveTimer = null;

function persistState(){
  try { localStorage.setItem(STATE_KEY, JSON.stringify(state)); }
  catch (_) {}
}

function queueSave(){
  if (saveTimer) return;
  saveTimer = setTimeout(() => { saveTimer = null; persistState(); }, 120);
}

function rememberScroll(el, key){
  if (!el) return;
  const saved = state.scroll[key];
  if (saved && typeof saved === 'object') {
    if (Number.isFinite(saved.top)) el.scrollTop = saved.top;
    if (Number.isFinite(saved.left)) el.scrollLeft = saved.left;
  } else if (Number.isFinite(saved)) {
    el.scrollTop = saved;
  }
  el.addEventListener('scroll', () => {
    const next = state.scroll[key];
    if (!next || typeof next !== 'object') {
      state.scroll[key] = {top: el.scrollTop, left: el.scrollLeft};
    } else {
      next.top = el.scrollTop;
      next.left = el.scrollLeft;
    }
    queueSave();
  }, {passive:true});
}

function rememberDetails(spanId, el){
  if (!spanId || !el) return;
  el.dataset.spanId = spanId;
  if (detailsOpen.has(spanId)) el.open = true;
  el.addEventListener('toggle', () => {
    if (el.open) detailsOpen.add(spanId); else detailsOpen.delete(spanId);
    state.detailsOpen = Array.from(detailsOpen);
    queueSave();
  });
}

window.addEventListener('beforeunload', persistState);

const RAW = /*__EMBED__*/;
// Build maps by span_id and group by trace_id, merging span_start + finished span
const spans = new Map();
const traces = new Map();
let warnCount = 0;

for (const rec of RAW) {
  if (rec.type === "span_start") {
    const s = spans.get(rec.span_id) || {};
    spans.set(rec.span_id, Object.assign(s, rec, {events: s.events || []}));
  } else {
    // finished span
    const s = spans.get(rec.span_id) || {};
    const merged = Object.assign(s, rec);
    spans.set(rec.span_id, merged);
    if (merged.events) {
      for (const e of merged.events) {
        if (e.name === "warning") warnCount++;
      }
    }
  }
}

// index by trace, also compute min/max times
for (const s of spans.values()) {
  const tid = s.trace_id;
  if (!traces.has(tid)) traces.set(tid, {spans:[], min:s.start_time_unix_nano||s.start||0, max:s.end_time_unix_nano||s.start_time_unix_nano||0});
  const t = traces.get(tid);
  t.spans.push(s);
  const st = s.start_time_unix_nano || s.start || 0;
  const en = s.end_time_unix_nano || st;
  if (st && st < t.min) t.min = st;
  if (en && en > t.max) t.max = en;
}

document.getElementById('c-traces').textContent = `${traces.size} traces`;
document.getElementById('c-spans').textContent = `${spans.size} spans`;
document.getElementById('c-warn').textContent  = `${warnCount} warnings`;

// Render tree + timelines per trace
const tree = document.getElementById('tree');
const timelines = document.getElementById('timelines');

function nsToMs(ns){ return ns ? (ns/1e6).toFixed(1) : "0.0"; }

for (const [tid, tinfo] of traces.entries()) {
  // build tree: parent -> children
  const kids = new Map();
  const nodes = new Map();
  let roots = [];
  for (const s of tinfo.spans) {
    nodes.set(s.span_id, s);
    if (s.parent_span_id) {
      if (!kids.has(s.parent_span_id)) kids.set(s.parent_span_id, []);
      kids.get(s.parent_span_id).push(s);
    } else {
      roots.push(s);
    }
  }

  // tree panel
  const sec = document.createElement('section');
  const h = document.createElement('h3');
  h.textContent = `trace ${tid.slice(0,8)}…`;
  h.style.padding = "8px 10px"; h.style.margin = "0"; h.style.borderBottom = "1px solid #1f2937";
  sec.appendChild(h);

  const ul = document.createElement('ul'); sec.appendChild(ul);

  function statusClass(s){
    if (!s.end_time_unix_nano) return "run";
    if (s.status === "ERROR") return "err";
    return "ok";
  }

  function liFor(s, depth=0){
    const li = document.createElement('li'); li.className = statusClass(s);
    const nm = document.createElement('div'); nm.className = 'name';
    nm.textContent = s.name || '(unnamed)';
    const meta = document.createElement('div'); meta.className = 'meta';
    const dur = (s.end_time_unix_nano ? (s.end_time_unix_nano - (s.start_time_unix_nano||0))
               : ((Date.now()*1e6) - (s.start_time_unix_nano||0)));
    meta.innerHTML = `span <code>${s.span_id.slice(0,8)}…</code> • ${nsToMs(dur)} ms • ${s.kind||'INTERNAL'}${s.status?(' • '+s.status):''}`;
    li.appendChild(nm); li.appendChild(meta);

    const evc = (s.events||[]).length;
    if (evc) {
      const det = document.createElement('details'); const sum = document.createElement('summary');
      sum.textContent = `${evc} event${evc>1?'s':''}`; det.appendChild(sum);
      const inner = document.createElement('div'); inner.style.padding="4px 8px";
      for (const e of s.events) {
        const line = document.createElement('div');
        line.innerHTML = `<code>${e.name}</code> @ ${nsToMs(e.ts - (s.start_time_unix_nano||0))}ms ${e.attributes?JSON.stringify(e.attributes):''}`;
        inner.appendChild(line);
      }
      det.appendChild(inner); li.appendChild(det);
      rememberDetails(s.span_id, det);
    }

    const ch = kids.get(s.span_id) || [];
    if (ch.length){
      const u = document.createElement('ul');
      for (const c of ch) u.appendChild(liFor(c, depth+1));
      li.appendChild(u);
    }
    return li;
  }

  for (const r of roots) ul.appendChild(liFor(r));

  tree.appendChild(sec);

  // timeline (simple horizontal bars, relative to trace window)
  const tl = document.createElement('section');
  const h2 = document.createElement('h3');
  h2.textContent = `timeline ${tid.slice(0,8)}…`;
  h2.style.padding = "8px 10px"; h2.style.margin = "0"; h2.style.borderBottom = "1px solid #1f2937";
  tl.appendChild(h2);

  const win = (tinfo.max - tinfo.min) || 1;
  for (const s of tinfo.spans) {
    const row = document.createElement('div'); row.className = `tl ${statusClass(s)}`;
    const st = (s.start_time_unix_nano||0) - tinfo.min;
    const en = (s.end_time_unix_nano||Date.now()*1e6) - tinfo.min;
    const left = (100*st/win); const width = Math.max(0.8, 100*(en-st)/win);
    const bar = document.createElement('div'); bar.className = 'bar';
    bar.style.left = left+"%"; bar.style.width = width+"%";
    row.appendChild(bar);
    const meta = document.createElement('div');
    meta.className = 'meta'; meta.style.position='absolute'; meta.style.left='8px'; meta.style.top='2px';
    meta.textContent = `${s.name} (${nsToMs(en-st)} ms)`;
    row.appendChild(meta);
    tl.appendChild(row);
  }

  timelines.appendChild(tl);
}

rememberScroll(tree, 'tree');
rememberScroll(timelines, 'timelines');

// running count
let open = 0;
for (const s of spans.values()) if (!s.end_time_unix_nano) open++;
document.getElementById('c-open').textContent = `${open} running`;
</script>
"""



from .ir import Node, Model, Capability, Engine
from .ir import Type, ScalarType, ListType, UnionType, Field, Relation
from .ir import Task, Logical, Union, Sequence, Match, Until, Wait
from .ir import Not, Exists, ForAll, Loop, Break
from .ir import Var, Literal, Value, Annotation, Effect, Update, Lookup, Output, Construct, Aggregate

__all__ = [
    "Node",
    "Model",
    "Capability",
    "Engine",
    "Type",
    "ScalarType",
    "ListType",
    "UnionType",
    "Field",
    "Relation",
    "Task",
    "Logical",
    "Union",
    "Sequence",
    "Match",
    "Until",
    "Wait",
    "Not",
    "Exists",
    "ForAll",
    "Loop",
    "Break",
    "Var",
    "Literal",
    "Value",
    "Annotation",
    "Effect",
    "Update",
    "Lookup",
    "Output",
    "Construct",
    "Aggregate"
]



from __future__ import annotations
from pandas import DataFrame
from .errors import InvalidVersionKind, RAIInvalidVersionWarning
from packaging.version import Version
from typing import Dict, List, Optional, Tuple

# A simple container for 2 version objects, a lower bound and an upper bound, plus a version
# range string.
#
# Equivalent to a packaging SpecifierSet defined as ">={lower_bound}, <{upper_bound}", but
# has the benefit that we can check whether a version that is not contained by the set is
# lower than the lower bound or higher than the higher bound, which is important for error
# reporting.
# The vrange string is the Rel package manager representation of the low-high range. It
# differs from the python notation in that the upper bounds use wildcards and are inclusive
# (i.e. the range is <={upper bound}).
def version_range(low:str, high:str, vrange:str) -> tuple[Version, Version, str]:
    return (Version(low), Version(high), vrange)

# A list of rel libraries that the current version of relationalai depends on, including the
# range of compatible versions.
#
# NOTE: these dependencies are only verified if the use_package_manager option is enabled.
#
PACKAGE_DEPENDENCIES = [
    ('std', '24c1872c-de8a-5b12-7648-3d72f007a7a9', version_range('0.1.9', '0.2.0', '0.1.9 - 0.1.*'))
]

# Dependencies to libraries that are not packages, so we only check that their versions are
# compatible, but do not attempt to maintain them. Eventually this will be removed.
STATIC_DEPENDENCIES = [
    ('graphlib', '04cd9c07-16aa-4433-bdd7-adfdfaee317d', version_range('0.3.0', '0.4.0', '0.3.0 - 0.3.*')),
    ('std', '24c1872c-de8a-5b12-7648-3d72f007a7a9', version_range('0.1.5', '0.2.0', '0.1.5 - 0.1.*'))
]

#--------------------------------------------------
# Public API
#--------------------------------------------------

#
# Generate a Rel query that will lookup the versions of Rel libraries currently installed in
# the database, as well as whether the registry is up to date. This is used if the package
# manager is enabled.
#
def generate_query_package_manager():
    return _generate_query_package_manager(PACKAGE_DEPENDENCIES, STATIC_DEPENDENCIES)

#
# Generate a Rel query that will just check the static dependency versions. This is used if
# the package manager is not enabled.
#
def generate_query_version_check():
    # do not check package dependencies
    return _generate_query_version_check([], STATIC_DEPENDENCIES)

#
# Analyze packages in the database.
#
# This is similar to check_package_manager, but will only check the static dependencies.
# This is used when the package manager is not enabled.
#
def check_static_dependencies(
    installed_dependencies: Dict,
    platform: str,
    app_name: str,
    engine_name: str,
    model_name: str,
    warn_on_packages: bool = False
) -> None:
    _compare_dependencies(
        package_deps=[],
        static_deps=STATIC_DEPENDENCIES,
        installed_dependencies=installed_dependencies,
        platform=platform,
        app_name=app_name,
        engine_name=engine_name,
        model_name=model_name,
        warn_on_packages=warn_on_packages
    )

#
# Analyze the response of the "query_version_check" transaction.
#
# This is similar to check_package_manager_fallback, but will only check the static
# dependencies. This is used when the package manager is not enabled.
#
def check_static_dependencies_fallback(
    response: DataFrame,
    platform: str,
    app_name: str,
    engine_name: str,
    model_name: str,
    warn_on_packages: bool = False,
) -> None:
    results = response.results
    _compare_dependencies_fallback(
        [],
        STATIC_DEPENDENCIES,
        _extract_library_versions(results),
        _extract_std_version(results),
        platform,
        app_name,
        engine_name,
        model_name,
        warn_on_packages
    )

#
# Analyze the installed packages in the database.
#
# Return a tuple of booleans. The first boolean represents whether the registry is outdated
# and need to be updated. The second boolean represents whether the packages in the
# database are outdated and need to be updated.
#
# If `warn_on_packages` is set, also warn on outdated packages.
def check_package_manager(
    installed_dependencies: Dict,
    platform: str,
    app_name: str,
    engine_name: str,
    model_name: str,
    warn_on_packages: bool = False,
) -> Tuple[bool, bool]:
    update_packages, update_registry = _compare_dependencies(
        package_deps=PACKAGE_DEPENDENCIES,
        static_deps=STATIC_DEPENDENCIES,
        installed_dependencies=installed_dependencies,
        platform=platform,
        app_name=app_name,
        engine_name=engine_name,
        model_name=model_name,
        warn_on_packages=warn_on_packages
    )
    return update_registry, update_packages

#
# Analyze the response of the "query_package_manager" transaction.
#
# Return a tuple of booleans. The first boolean represents whether the registry is outdated
# and need to be updated. The second boolean represents whether the packages in the
# database are outdated and need to be updated.
#
# If `warn_on_packages` is set, also warn on outdated packages.
def check_package_manager_fallback(
    response: DataFrame,
    platform: str,
    app_name: str,
    engine_name: str,
    model_name: str,
    warn_on_packages: bool = False,
) -> Tuple[bool, bool]:
    results = response.results
    update_registry = _extract_update_registry(results)
    # if we need to update the registry, we also need to update packages
    if update_registry:
        return True, True

    update_packages = _compare_dependencies_fallback(
        PACKAGE_DEPENDENCIES,
        STATIC_DEPENDENCIES,
        _extract_library_versions(results),
        _extract_std_version(results),
        platform,
        app_name,
        engine_name,
        model_name,
        warn_on_packages
    )
    return False, update_packages

#
# Generate a Rel query to update the package registry, i.e. update the metadata to ensure
# the database is up to date with the available packages.
#
def generate_update_registry():
    return '''
    def response { ::std::pkg::registry::update_by_name["RAI"]}
    def insert { response[:insert] }
    def delete { response[:delete] }
    '''

#
# Generate a Rel query to update the package dependencies in the database to the latest
# available version.
#
def generate_update_packages():
    return _generate_update_packages(PACKAGE_DEPENDENCIES)

#--------------------------------------------------
# Implementation Details
#--------------------------------------------------

def _extract_update_registry(results):
    for result in results:
        if result["relationId"].startswith("/:output/:update_registry"):
            return True
    return False

def _extract_std_version(results):
    for result in results:
        if result["relationId"].startswith("/:output/:std"):
            for (version,) in result["table"].itertuples(index=False):
                return version
    return None

def _extract_library_versions(results):
    libraries = {}
    for result in results:
        if result["relationId"].startswith("/:output/:static_lock"):
            for (name, uuid, version) in result["table"].itertuples(index=False):
                libraries[(name, uuid)] = version
    return libraries

# Parse the package_versions coming from ERP metadata into the old format used by the
# fallback version check via the engine.
def _parse_package_versions_to_lock(package_versions: Dict[str, Dict[str, str]]) -> Dict[Tuple[str, str], str]:
    return {(version_dict["name"], uuid): version_dict["version"] for uuid, version_dict in package_versions.items()}

# Compare the current packages against the expected package_deps and static_deps. Raise
# warnings when necessary and return whether we should make the package manager update the
# package versions or even update the registry.
#
# Returns (update_packages, update_registry)
def _compare_dependencies(
        package_deps: List,
        static_deps: List,
        installed_dependencies: Dict,
        platform: str,
        app_name: str,
        engine_name: str,
        model_name: str,
        warn_on_packages: bool = False
    ) -> Tuple[bool, bool]:
    lock = _parse_package_versions_to_lock(installed_dependencies)

    # dependencies not yet kept in erp metadata. We need to fallback to retrieving package
    # version data from the engine directly.
    if not installed_dependencies:
        return True, True

    _warn_on_dependency_mismatch(
        package_deps=package_deps, 
        static_deps=static_deps, 
        lock=lock, 
        std=None,
        platform=platform, 
        app_name=app_name, 
        engine_name=engine_name, 
        model_name=model_name, 
        warn_on_packages=warn_on_packages,
    )

    # finally, check that all package dependencies are satisfied
    for (_, uuid, vrange) in package_deps:
        # library is missing from the database
        if uuid not in installed_dependencies:
            # currently, we do not know if the package is available in the registry, as the
            # ERP only stores information about install packages. So we need to update it to
            # make sure it exists
            return True, True

        version = Version(installed_dependencies[uuid]["version"])

        # version < vrange[0] -> Rel library version lower than lower bound, too old
        if version < vrange[0]:
            # if the installed version is too old, check if the fitting one is available
            # in the registry
            max_available_version = Version(installed_dependencies[uuid]["max_version"])
            if max_available_version < vrange[0]:
                return True, True
            return True, False

        # version >= vrange[1] -> Rel library version higher than upper bound, too new
        if version >= vrange[1]:
            return True, False

    # all dependencies satisfied
    return False, False

# Compare the current packages (in lock) and the value of std::version (in std) against the
# expected package_deps and static_deps. Raise warnings when necessary and return whether we
# should make the package manager update the package versions. This method is a fallback for
# when package metadata is not available in ERP and was instead retrieved by querying the
# engine. The engine response has a different structure than the ERP response, so we need to
# handle it differently.
def _compare_dependencies_fallback(
        package_deps: List,
        static_deps: List,
        lock: Dict,
        std: Optional[str],
        platform: str,
        app_name: str,
        engine_name: str,
        model_name: str,
        warn_on_packages = False,
    ) -> bool:
    _warn_on_dependency_mismatch(
        package_deps=package_deps, 
        static_deps=static_deps, 
        lock=lock, 
        std=std, 
        platform=platform, 
        app_name=app_name, 
        engine_name=engine_name, 
        model_name=model_name, 
        warn_on_packages=warn_on_packages,
    )

    # database is pre-versioning (before 0.1.0 was defined)
    if not lock and not std:
        return False

    # database is pre-package manager (between 0.1.0 and 0.1.5)
    if std:
        return False

    # finally, check that all package dependencies are satisfied
    for (name, uuid, vrange) in package_deps:
        key = (name, uuid)
        # library is missing from the database
        if key not in lock:
            return True

        version = Version(lock[key])
        # version >= vrange[1] -> Rel library version higher than upper bound, too new
        # version < vrange[0] -> Rel library version lower than lower bound, too old
        if version >= vrange[1] or version < vrange[0]:
            return True
    # all dependencies satisfied
    return False

# Check the static dependencies and potentially raise warnings. If `warn_on_packages` is
# set, also check and raise on package dependencies. Used for version checking without
# intent to upgrade.
def _warn_on_dependency_mismatch(
        package_deps: List,
        static_deps: List,
        lock: Dict,
        std: Optional[str],
        platform: str,
        app_name: str,
        engine_name: str,
        model_name: str,
        warn_on_packages = False,
    ) -> None:
    def warn(kind: InvalidVersionKind, expected, errors: list[tuple[InvalidVersionKind, str, str]] = []):
        RAIInvalidVersionWarning(
            kind=kind,
            expected=expected,
            lock=lock,
            platform=platform,
            app_name=app_name,
            engine_name=engine_name,
            model_name=model_name,
            errors=errors
        )
        return False

    # database is pre-versioning (before 0.1.0 was defined)
    if not lock and not std:
        warn(InvalidVersionKind.SchemaOutOfDate, static_deps)
        return

    # database is pre-package manager (between 0.1.0 and 0.1.5)
    if std:
        warn(InvalidVersionKind.SchemaOutOfDate, static_deps)
        return

    # database has a package manager, first, check the static dependencies and potentially
    # raise warnings
    # If `warn_on_packages` is set, also check and raise on package dependencies.
    # Used for version checking without intent to upgrade.
    deps_to_error_on = static_deps + package_deps if warn_on_packages else static_deps
    errors = []
    for (name, uuid, vrange) in deps_to_error_on:
        key = (name, uuid)
        # library is missing from the database
        if key not in lock:
            errors.append((InvalidVersionKind.SchemaOutOfDate, name, uuid))
        else:
            version = Version(lock[key])
            # Rel library version higher than upper bound, too new
            if version >= vrange[1]:
                errors.append((InvalidVersionKind.LibraryOutOfDate, name, uuid))

            # Rel library version lower than lower bound, too old
            if version < vrange[0]:
                errors.append((InvalidVersionKind.SchemaOutOfDate, name, uuid))

    if errors:
        if all(kind == InvalidVersionKind.LibraryOutOfDate for (kind, _, _) in errors):
            kind = InvalidVersionKind.LibraryOutOfDate
        elif all(kind == InvalidVersionKind.SchemaOutOfDate for (kind, _, _) in errors):
            kind = InvalidVersionKind.SchemaOutOfDate
        else:
            kind = InvalidVersionKind.Incompatible

        warn(kind, deps_to_error_on, errors)

def _generate_query_version_check(package_deps, static_deps):
    # deduplicate package entries
    checks = dict((name, uuid) for name, uuid, _ in package_deps)
    for name, uuid, _ in static_deps:
        if name not in checks:
            checks[name] = uuid

    static_lock_bindings = " ;\n            ".join(
        ["(\"%s\", \"%s\")" % (name, uuid) for name, uuid in checks.items()]
    )

    return f'''
    @no_diagnostics(:UNDEFINED_IDENTIFIER)
    def output[:std]: {{ std::version }}

    @no_diagnostics(:TYPE_MISMATCH)
    def output(:static_lock, name, uuid, version):
        rel(:pkg, :std, :pkg, :project, :static_lock, name, uuid, version) and
        {{
            {static_lock_bindings}
        }}(name, uuid)'''

def _generate_query_package_manager(package_deps, static_deps):
    lower_bounds = " ;\n        ".join(
        ["(\"%s\", \"%s\")" % (uuid, vrange[0]) for _, uuid, vrange in package_deps]
    )

    return f'''
    {_generate_query_version_check(package_deps, static_deps)}

    @no_diagnostics(:UNDEFINED_IDENTIFIER)
    def output[:update_registry]:
        exists((uuid, version) |
            lower_bounds(uuid, version) and
            not available(uuid, version) and
            ::std::pkg::registry::uuid(_, _)
        )

    def lower_bounds {{
        {lower_bounds}
    }}

    @no_diagnostics(:UNDEFINED_IDENTIFIER)
    @inline
    def available(uuid, version): {{
        exists((pkg, v, pv) |
            ::std::pkg::^Package(uuid_from_string[uuid], pkg) and
            ::std::pkg::^Version(version, v) and
            ::std::pkg::^PackageVersion(pkg, v, pv) and
            ::std::pkg::package_version::version(pv, v)
        )
    }}'''


def _generate_update_packages(package_deps):
    bindings = " ;\n        ".join(
        ["\"%s@%s\"" % (name, vrange[2]) for name, _, vrange in package_deps]
    )

    return f'''
    def response {{ ::std::pkg::project::update_package[{{
        {bindings}
    }}]}}
    def insert {{ response[:insert] }}
    def delete {{ response[:delete] }}
    '''



from v0.relationalai.early_access.dsl.core.relations import AbstractRelationSignature, RelationSignature
from v0.relationalai.early_access.dsl.core.types import Type
from v0.relationalai.early_access.dsl.core.types.standard import PositiveInteger, UnsignedInteger

class AbstractFunctionSignature(AbstractRelationSignature):

    def functional(self): return True

class EmptyRelation:

    def __init__(self, sig):
        self._signature = sig

    def display(self): return 'EMPTYSET'

    def entityid(self): return hash((EmptyRelation, self._signature.entityid()))

    def pprint(self):
        return self.display() + self._signature.display()

    def rename(self, bdgs): return self

class FunctionSignature(AbstractFunctionSignature):
    pass

class Sequence(AbstractFunctionSignature):

    def __init__(self, T, *types):
        super().__init__(*types)
        self._range = T

    def range(self): return self._range

class SequenceOne(Sequence):

    def __init__(self, T):
        super().__init__(T, PositiveInteger, T)

    def display(self):
        return f"seq {self.range().display()}"

    def entityid(self):
        return hash((Sequence, self.range().entityid()))

class SequenceZero(Sequence):

    def __init__(self, T):
        super().__init__(T, UnsignedInteger, T)

    def display(self):
        return f"seq0 {self.range().display()}"

    def entityid(self):
        return hash((SequenceZero, self.range().entityid()))

class InjectiveSequenceOne(SequenceOne):

    def __init__(self, T):
        Sequence.__init__(self, T)

    def display(self):
        return f"iseq {self.range().display()}"

    def entityid(self):
        return hash((InjectiveSequenceOne, self.range().entityid()))


class InjectiveSequenceZero(SequenceZero):

    def __init__(self, T):
        Sequence.__init__(self, T)

    def display(self):
        return f"iseq0 {self.range().display()}"

    def entityid(self):
        return hash((InjectiveSequenceZero, self.range().entityid()))

class FunctionFactory:

    def __call__(self, *args):

        nargs = len(args)

        if nargs < 2 or nargs > 3:
            raise Exception("Must instantiate functions with two type arguments and an optional name")

        dtype = args[0]
        rtype = args[1]

        if not isinstance(dtype, Type):
            raise Exception(f"Function signature domain must be a valid Type, not {dtype}")

        if not isinstance(rtype, Type):
            raise Exception(f"Function signature range must be a valid Type, not {rtype}")

        rsig = FunctionSignature((dtype, rtype))

        if nargs == 3:
            name = args[2]
            if isinstance(name, str):
                return rsig(name)
            else:
                raise Exception("Component name provide when instantiating Func is not a string")
        else:
            return rsig


class RelationFactory:

    def __call__(self, *args):

        if len(args) == 0:
            raise Exception("Must instantiate relations with one or more column types and an optional name")

        nargs = len(args)
        includes_name = isinstance(args[nargs - 1], str)
        columns = nargs - 1 if includes_name else nargs
        column_types = []
        for i in range(columns):
            if not isinstance(args[i], Type):
                raise Exception(f"Relation signatures must comprise a types, not {args[i]}")

            column_types.append(args[i])
        rsig = RelationSignature(*column_types)

        if includes_name:
            return rsig(args[nargs - 1])
        else:
            return rsig


class SequenceFactory:

    def __init__(self, ctor):
        self._ctor = ctor

    def __call__(self, *args):

        if len(args) == 0:
            raise Exception("Must instantiate sequence types with an element type")

        if len(args) > 2:
            raise Exception("Cannot create sequences of composite element type")

        element_type = args[0]
        if not isinstance(element_type, Type):
            raise Exception("First parameter of Sequence instance must be an element type")

        seq = self._ctor(element_type)
        if len(args) == 2:
            if isinstance(args[1], str):
                return seq(args[1])
            else:
                raise Exception("Can only instantiate sequence type with a string name")
        else:
            return seq


# Constructors that can be used in Schema declarations for relations of
# various kinds that are not declared in an ontology.
#

Rel = RelationFactory()
Func = FunctionFactory()

Seq = SequenceFactory(SequenceOne)
Seq0 = SequenceFactory(SequenceZero)
Iseq = SequenceFactory(InjectiveSequenceOne)
Iseq0 = SequenceFactory(InjectiveSequenceZero)



"""
Main configuration class for PyRel (YAML-based).

"""

from __future__ import annotations

from abc import ABC
from typing import Any, TypeVar, overload, Literal, TYPE_CHECKING
from pydantic import Field, model_validator
from pydantic_settings import SettingsConfigDict

try:
    from confocal import BaseConfig
except ImportError:
    # Confocal not yet published to PyPI - use pydantic_settings as fallback
    # Config system not actively used yet, this is just to prevent import errors
    from pydantic_settings import BaseSettings as BaseConfig  # type: ignore[misc, assignment]

if TYPE_CHECKING:
    import snowflake.snowpark
    import duckdb

from .connections import ConnectionConfig, BaseConnection, SnowflakeConnection, DuckDBConnection
from .config_fields import EngineConfig, DataConfig, CompilerConfig, ModelConfig, ReasonerConfig, DebugConfig
from .external.dbt_converter import convert_dbt_to_rai
from .external.snowflake_converter import convert_snowflake_to_rai
from .external.utils import find_dbt_profiles_file, find_snowflake_config_file

# TypeVar for generic connection retrieval
T = TypeVar('T', bound=BaseConnection)


class _Config(BaseConfig, ABC): # type: ignore for now until we publish confocal
    """Base configuration class with common fields and methods."""

    active_profile: str | None = Field(
        default=None,
        description="Currently active profile name (auto-defaults if only one profile)"
    )

    profiles: dict[str, dict[str, Any]] = Field(
        default_factory=dict,
        alias="profile",
        description="Profiles available (dev, prod, staging, etc.) - profiles overlay config fields"
    )

    connections: dict[str, ConnectionConfig] = Field(
        default_factory=dict,
        description="Connections available (snowflake, duckdb, etc.)"
    )

    default_connection: str | None = Field(
        default=None,
        description="Default connection name (auto-defaults if only one connection)"
    )

    use_graph_index: bool = Field(
        default=True,
        description="Enabling graph index"
    )
    use_direct_access: bool = Field(
        default=False,
        description="Use direct access mode"
    )
    describe_optimization: str | None = Field(default=None, description="Describe optimization")
    enable_otel_handler: bool = Field(
        default=False,
        description="Enable OpenTelemetry handler"
    )

    # Nested config objects
    engine: EngineConfig = Field(
        default_factory=EngineConfig,
        description="Engine configuration"
    )
    data: DataConfig = Field(
        default_factory=DataConfig,
        description="Data loading and streaming configuration"
    )
    compiler: CompilerConfig = Field(
        default_factory=CompilerConfig,
        description="Compiler configuration"
    )
    model: ModelConfig = Field(
        default_factory=ModelConfig,
        description="Model configuration"
    )
    reasoner: ReasonerConfig = Field(
        default_factory=ReasonerConfig,
        description="Reasoner configuration"
    )
    debug: DebugConfig = Field(
        default_factory=DebugConfig,
        description="Debug configuration"
    )

    @model_validator(mode='before')
    @classmethod
    def set_default_snowflake_authenticator(cls, data: Any):
        """Set default authenticator for Snowflake connections in raw data."""
        if not isinstance(data, dict):
            return data

        def process_connections(connections):
            if not isinstance(connections, dict):
                return
            for conn_data in connections.values():
                if isinstance(conn_data, dict) and conn_data.get('type') == 'snowflake':
                    conn_data.setdefault('authenticator', 'username_password')

        # Process top-level connections
        if 'connections' in data:
            process_connections(data['connections'])

        # Process connections in profiles (before overlay)
        if 'profiles' in data and isinstance(data['profiles'], dict):
            for profile_data in data['profiles'].values():
                if isinstance(profile_data, dict) and 'connections' in profile_data:
                    process_connections(profile_data['connections'])

        return data

    @model_validator(mode='after')
    def validate_connections(self):
        if not self.connections:
            raise ValueError("Config must have at least one connection")

        # Auto-set default_connection if only one connection exists and not already set
        if len(self.connections) == 1 and self.default_connection is None:
            self.default_connection = next(iter(self.connections.keys()))

        return self

    def get_default_connection(self) -> ConnectionConfig:
        default_conn_name = self.default_connection

        if default_conn_name is None:
            if len(self.connections) == 1:
                default_conn_name = list(self.connections.keys())[0]
            else:
                raise ValueError(
                    f"Multiple connections available but no default_connection specified. "
                    f"Available: {list(self.connections.keys())}"
                )

        if default_conn_name not in self.connections:
            raise ValueError(
                f"default_connection '{default_conn_name}' not found. "
                f"Available: {list(self.connections.keys())}"
            )

        return self.connections[default_conn_name]

    def get_connection(self, connection_type: type[T], name: str | None = None) -> T:
        if name is None:
            connection = self.get_default_connection()
        else:
            if name not in self.connections:
                raise ValueError(
                    f"Connection '{name}' not found. "
                    f"Available: {list(self.connections.keys())}"
                )
            connection = self.connections[name]

        if not isinstance(connection, connection_type):
            # Provide helpful error message with the actual connection class hierarchy
            actual_type = type(connection).__name__
            # Check if it's a Snowflake connection
            if isinstance(connection, SnowflakeConnection):
                connection_class = "SnowflakeConnection"
            elif isinstance(connection, DuckDBConnection):
                connection_class = "DuckDBConnection"
            else:
                connection_class = actual_type

            raise ValueError(
                f"Connection '{name or self.default_connection}' is not of type {connection_type.__name__}. "
                f"Got: {connection_class} (actual: {actual_type})"
            )

        return connection

    @overload
    def get_session(self, connection_type: type[SnowflakeConnection]) -> snowflake.snowpark.Session: ...

    @overload
    def get_session(self, connection_type: type[DuckDBConnection]) -> duckdb.DuckDBPyConnection: ...

    @overload
    def get_session(self, connection_type: None = None) -> snowflake.snowpark.Session | duckdb.DuckDBPyConnection: ...

    def get_session(self, connection_type: type[SnowflakeConnection] | type[DuckDBConnection] | None = None) -> snowflake.snowpark.Session | duckdb.DuckDBPyConnection:
        if connection_type is None:
            connection = self.get_default_connection()
            return connection.get_session()
        else:
            connection = self.get_connection(connection_type)
            return connection.get_session()


# =============================================================================
# Config Source Classes
# =============================================================================

class RAIConfig(_Config):
    """Config loaded from raiconfig.yaml."""

    source: Literal["rai"] = Field(default="rai", exclude=True)

    model_config = SettingsConfigDict(
        yaml_file="raiconfig.yaml",
        extra="ignore",
        nested_model_default_partial_update=True,
    )


class ConfigFromSnowflake(_Config):
    source: Literal["snowflake"] = Field(default="snowflake", exclude=True)

    model_config = SettingsConfigDict(
        toml_file=find_snowflake_config_file(),
        extra="ignore",
        nested_model_default_partial_update=True,
    )

    @model_validator(mode='before')
    @classmethod
    def convert_snowflake_structure(cls, data: Any):
        if not isinstance(data, dict):
            return data
        return convert_snowflake_to_rai(data)

class ConfigFromDBT(_Config):
    source: Literal["dbt"] = Field(default="dbt", exclude=True)

    model_config = SettingsConfigDict(
        yaml_file=find_dbt_profiles_file(),
        extra="ignore",
        nested_model_default_partial_update=True,
    )

    @model_validator(mode='before')
    @classmethod
    def convert_dbt_structure(cls, data: Any):
        """Convert DBT profiles.yml format to RAI Config format."""
        if not isinstance(data, dict):
            return data

        return convert_dbt_to_rai(data)


# =============================================================================
# Config Factory - Tries each source in order
# =============================================================================

def Config(**data) -> _Config:
    """
    Create Config by trying multiple sources in priority order:
    1. RAIConfig (raiconfig.yaml) - or direct data if provided
    2. ConfigFromSnowflake (config.toml) - only if no data provided
    3. ConfigFromDBT (profiles.yml) - only if no data provided
    """
    sources = [
        ("RAIConfig (raiconfig.yaml)", RAIConfig, True),      # passes **data
        ("ConfigFromSnowflake (config.toml)", ConfigFromSnowflake, False),
        ("ConfigFromDBT (profiles.yml)", ConfigFromDBT, False),
    ]

    errors = []

    for name, source_cls, accepts_data in sources:
        try:
            if accepts_data:
                return source_cls(**data)
            else:
                return source_cls()
        except Exception as e:
            errors.append((name, str(e)))

    # All failed, create helpful error message
    error_lines = [f"  ❌ {name}: {error}" for name, error in errors]
    error_summary = "\n".join(error_lines)

    raise FileNotFoundError(
        f"Could not load config from any source:\n\n{error_summary}\n\n"
        f"To fix: Correct one of the config files above, or pass config programmatically:\n"
        f"   Config(connections={{...}})"
    )



"""Base connection protocol for all connection types."""

from __future__ import annotations

from abc import ABC
from typing import Any
from pydantic import BaseModel, ConfigDict, PrivateAttr


class BaseConnection(BaseModel, ABC):
    # Note: `type` field is NOT defined here, subclasses must define it
    # with specific Literal types for discriminated unions to work correctly

    model_config = ConfigDict(arbitrary_types_allowed=True)

    # Private field for caching session (excluded from serialization)
    _cached_session: Any = PrivateAttr(default=None)

    def get_session(self) -> Any:
        raise NotImplementedError("Subclasses must implement get_session()")

    def clear_session_cache(self) -> None:
        self._cached_session = None



import warnings

from v0.relationalai.semantics.std.math import abs, isnan, isinf, maximum, natural_log, sqrt, minimum, ceil, floor

__all__ = ["abs", "isnan", "isinf", "maximum", "natural_log", "sqrt", "minimum", "ceil", "floor"]

warnings.warn(
    "relationalai.early_access.builder.std.math is deprecated, "
    "Please migrate to relationalai.semantics.std.math",
    DeprecationWarning,
    stacklevel=2,
)


from __future__ import annotations
from typing import Union
import textwrap
import uuid
import time

from v0.relationalai.semantics.snowflake import Table
from v0.relationalai.semantics import std
from v0.relationalai.semantics.internal import internal as b
from v0.relationalai.semantics.rel.executor import RelExecutor
from v0.relationalai.semantics.lqp.executor import LQPExecutor
from v0.relationalai.tools.constants import DEFAULT_QUERY_TIMEOUT_MINS
from v0.relationalai.util.timeout import calc_remaining_timeout_minutes

from .common import make_name
from v0.relationalai.experimental.solvers import Solver

_Any = Union[b.Producer, str, float, int]
_String = Union[b.Producer, str]
_Number = Union[b.Producer, float, int]

class SolverModelDev:
    def __init__(self, model: b.Model, num_type: str):
        self._model = model
        assert num_type in ("cont", "int")
        data_type = "int" if num_type == "int" else "float"
        self._data_type = data_type
        self._variable_relationships = set()
        self._expr_id = 0
        Relationship = model.Relationship

        # TODO workaround for QB not distinguishing string names for Concepts/Relationships, only a problem with raw Rel for result extraction maybe
        sm_id = next(b._global_id)
        def _name(s: str) -> str:
            return f"{s}_{sm_id}"

        # variable relations
        self.variables = Relationship("{node} is variable", short_name=_name("variables"))
        self.variable_type = Relationship("{node} has {type:str}", short_name=_name("variable_type")) # TODO use codes/enum
        self.variable_name = Relationship("{node} has {name:str}", short_name=_name("variable_name"))
        self.variable_start = Relationship(f"{{node}} has {{start:{data_type}}}", short_name=_name("variable_start"))

        # expression relations
        self.operator = Relationship("{node} has {op:str}", short_name=_name("operator")) # TODO use codes/enum
        self.ordered_args_hash = Relationship("{node} arg {i:int} is {arg}", short_name=_name("ordered_args_hash"))
        self.ordered_args_data = Relationship(f"{{node}} arg {{i:int}} is {{arg:{data_type}}}", short_name=_name("ordered_args_data"))
        self.unordered_args_hash = Relationship("{node} arg {i} is {arg}", short_name=_name("unordered_args_hash"))
        self.min_objectives = Relationship("{node} is min objective", short_name=_name("min_objectives"))
        self.max_objectives = Relationship("{node} is max objective", short_name=_name("max_objectives"))
        self.constraints = Relationship("{node} is constraint", short_name=_name("constraints"))
        self.expression_name = Relationship("{node} has {name:str}", short_name=_name("expression_name")) # TODO use codes/enum

        # solver result relations
        self.result_info = model.Relationship("{key:str} has {val:str}", short_name=_name("result_info"))
        self.point = model.Relationship(f"{{var}} has {{point:{data_type}}}", short_name=_name("point"))
        self.points = model.Relationship(f"point {{i:int}} for {{var}} has {{point:{data_type}}}", short_name=_name("points"))
        # self.objective_values = model.Relationship(f"point {{i:int}} has objective value {{val:{data_type}}}", short_name=_name("objective_values"))
        # self.primal_statuses = model.Relationship("point {i:int} has primal status {status:str}", short_name=_name("primal_statuses"))

        self._model_info = {
            "num_variables": self.variables,
            "num_constraints": self.constraints,
            "num_min_objectives": self.min_objectives,
            "num_max_objectives": self.max_objectives,
        }

    # TODO(coey) assert that it is a property? not just a relationship.
    def solve_for(self, expr, where: list = [], populate: bool = True, **kwargs):
        if isinstance(expr, b.Fragment):
            # TODO(coey) remove in future
            raise ValueError("select fragment argument to `solve_for` is deprecated; instead use `where = [conditions...]` kwarg to specify optional grounding conditions")
        elif isinstance(expr, b.Expression):
            # must be of the form rel(a, ..., x) where the last element is the decision variable
            rel = expr._op
            assert isinstance(rel, b.Relationship)
            params = expr._params
        elif isinstance(expr, b.Relationship):
            rel = expr
            params = [b.field_to_type(self._model, f) for f in rel._fields]
        else:
            raise ValueError(f"Invalid expression type {type(expr)} for `solve_for`")
        assert len(params) == len(rel._fields)
        assert rel not in self._variable_relationships

        self._variable_relationships.add(rel)
        node = b.Hash.ref()
        defs = [self.variables(node)]

        # handle optional variable properties
        new_kwargs = kwargs.copy()
        if "type" not in new_kwargs:
            new_kwargs["type"] = "cont" if self._data_type == "float" else "int"
        for (key, val) in new_kwargs.items():
            if key == "name":
                assert isinstance(val, _Any) or isinstance(val, list), f"Expected {key} to be a value or list, got {type(val)}"
                defs.append(self.variable_name(node, make_name(val)))
            elif key == "type":
                assert val in ("cont", "int", "bin"), f"Unsupported variable type {val} for `solve_for`; must be cont, int, or bin"
                defs.append(self.variable_type(node, val))
            elif key in ("lower", "upper", "fixed", "start"):
                assert isinstance(val, _Number), f"Expected {key} to be a number, got {type(val)}"
                if key == "start":
                    defs.append(self.variable_start(node, val))
                else:
                    # TODO use lower/upper/fixed tables in future rather than making constraints
                    op = ">=" if key == "lower" else ("<=" if key == "upper" else "=")
                    self.satisfy(b.require(b.Expression(b.Relationship.builtins[op], rel, val)).where(*where))
            else:
                raise ValueError(f"Invalid keyword argument {key} for `solve_for`")
        b.define(*defs).where(*where, _make_hash((rel._short_name, rel._parent or 0), node))

        if populate:
            # get variable values from the result point (populated by the solver)
            # TODO maybe instead delete/insert into variable relationships after solve.
            val = b.Number.ref()
            b.define(rel(val)).where(self.point(node, val), *where)

        return None

    def minimize(self, expr, name: _String | list | None = None):
        return self._handle_expr(self.min_objectives, expr, name)

    def maximize(self, expr, name: _String | list | None = None):
        return self._handle_expr(self.max_objectives, expr, name)

    def satisfy(self, expr: b.Fragment, check: bool = False, name: _String | list | None = None):
        assert expr._require, "Fragment input for `satisfy` must have a require clause"
        assert not expr._select and not expr._define, "Fragment input for `satisfy` must not have a select or define clause"
        if not check:
            # remove the `require` from the model roots so it is not checked
            b._remove_roots([expr])
        for req in expr._require:
            self._handle_expr(self.constraints, req, name, where=expr._where)
        return None

    def _handle_expr(self, root_type, expr, name: _String | list | None, where: list = []):
        ctx = ExprContext(self)
        ctx.where.extend(where)
        root = _rewrite(expr, ctx)
        assert root is not None, f"Cannot symbolify {expr}"
        ctx.define.append(root_type(root))
        # if name is not None:
        #     # TODO not printing expression names yet, so skip for now
        #     assert isinstance(name, (_String, list)), f"Expected name to be a string or list, got {type(name)}"
        #     ctx.define.append(self.expression_name(root, make_name(name)))
        self._handle_ctx(ctx)
        return None

    def _handle_ctx(self, ctx: ExprContext, where: list = []):
        all_where = ctx.where + where
        b.define(*ctx.define).where(*all_where)
        for subctx in ctx.subcontext:
            self._handle_ctx(subctx, where=all_where)

    # print the variables and components of the model in human-readable format
    def print(self, verbose: bool = True, expr_names: bool = False):
        # summarize variables
        var_types = ["int", "bin", "cont"]
        vi = b.Hash.ref()
        var_counts = b.select(*[b.count(vi).where(self.variable_type(vi, t)) | 0 for t in var_types]).to_df()
        assert var_counts.shape == (1, 3)
        (int_count, bin_count, cont_count) = var_counts.iloc[0]

        # summarize expressions
        expr_types = [
            (self.min_objectives, "Minimization objectives"),
            (self.max_objectives, "Maximization objectives"),
            (self.constraints, "Constraints"),
        ]
        ei = b.Hash.ref()
        expr_counts = b.select(*[b.count(ei).where(E(ei)) | 0 for (E, _) in expr_types]).to_df()
        assert expr_counts.shape == (1, 3)
        (min_count, max_count, cons_count) = expr_counts.iloc[0]

        # print summary
        print("Solver model has:")
        print(f"• {int_count} integer variables, {bin_count} binary variables, {cont_count} continuous variables")
        print(f"• {min_count} minimization objectives, {max_count} maximization objectives, {cons_count} constraints")
        if not verbose:
            return None

        # print variable names
        vn, vt = b.String.ref(), b.String.ref()
        var_names = b.select(vn.alias("name"), vt.alias("type")).where(self.variable_name(vi, vn), self.variable_type(vi, vt)).to_df()
        assert not var_names.empty, "No variable names found in the model"
        print("Variables:")
        print(var_names.to_string(index=False, header=True))

        # print expressions TODO print expression names optionally
        for (E, s) in expr_types:
            expr_strings = self._expr_strings(E)
            if expr_strings:
                print(s + ":")
                print("\n".join(expr_strings))
        return None

    def _expr_strings(self, roots):
        # materialize data
        iv = b.Hash.ref()
        roots_df = b.select(iv).where(roots(iv)).to_df()
        if roots_df.empty:
            return []
        nodes = roots_df.iloc[:, 0].tolist()

        sv = b.String.ref()
        names_dict = {k: v for k, v in b.select(iv, sv).where(self.variable_name(iv, sv))}
        ops_dict = {k: v for k, v in b.select(iv, sv).where(self.operator(iv, sv))}

        args_dict = {}
        iv2 = b.Hash.ref()
        iv3 = b.Hash.ref()
        # unordered args
        for k, _, v in b.select(iv, iv2, iv3).where(self.unordered_args_hash(iv, iv2, iv3)):
            if k not in args_dict:
                args_dict[k] = []
            args_dict[k].append(v)
        ordered_args_dict = {}
        iv2 = b.Integer.ref()
        for k, i, v in b.select(iv, iv2, iv3).where(self.ordered_args_hash(iv, iv2, iv3)):
            if k not in ordered_args_dict:
                ordered_args_dict[k] = []
            ordered_args_dict[k].append((i, v))
        vv = b.python_types_str_to_concepts[self._data_type].ref()
        for k, i, v in b.select(iv, iv2, vv).where(self.ordered_args_data(iv, iv2, vv)):
            if k not in ordered_args_dict:
                ordered_args_dict[k] = []
            ordered_args_dict[k].append((i, v))
        # convert ordered args to dict
        for k, t in ordered_args_dict.items():
            # sort by index
            t.sort(key=lambda x: x[0])
            args_dict[k] = [v for (_, v) in t]

        # compute strings
        return sorted(_expr_strings_rec(n, names_dict, ops_dict, args_dict) for n in nodes)

    # solve the model given a solver and solver options
    def solve(self, solver: Solver, log_to_console: bool = False, **kwargs):
        # validate options
        for k, v in kwargs.items():
            if not isinstance(k, str):
                raise ValueError(f"Invalid parameter key. Expected string, got {type(k)} for {k}.")
            if not isinstance(v, (int, float, str, bool)):
                raise ValueError(
                    f"Invalid parameter value. Expected string, integer, float, or boolean, got {type(v)} for {k}."
                )

        # set up
        input_id = uuid.uuid4().hex.upper()
        executor = self._model._to_executor()
        assert isinstance(executor, (RelExecutor, LQPExecutor))
        resources = executor.resources
        app_name = resources.get_app_name()
        print(app_name)

        # Note: currently the query timeout is not propagated to the steps 'export model
        # relations', and 'import result relations'. For those steps the default query
        # timeout value defined in the config will apply.
        # TODO: propagate the query timeout to those steps as well.
        query_timeout_mins = kwargs.get("query_timeout_mins", None)
        config = self._model._config
        if query_timeout_mins is None and (timeout_value := config.get("query_timeout_mins", DEFAULT_QUERY_TIMEOUT_MINS)) is not None:
            query_timeout_mins = int(timeout_value)
        config_file_path = getattr(config, 'file_path', None)
        start_time = time.monotonic()

        # 1. export model relations
        print("export model relations")
        # TODO(coey) perf: only export the relations that are actually used in the model
        to_export = [
            "variables",
            "variable_type",
            # "variable_name",
            # "operator",
            # "ordered_args_hash",
            # "ordered_args_data",
            # "unordered_args_hash",
            # "min_objectives",
            # "max_objectives",
            # "constraints",
        ]
        for name in to_export:
            print(f"exporting {name}")
            rel = getattr(self, name)
            table_name = f"{app_name}.RESULTS.SOLVER_{input_id}_{name.upper()}"
            table = Table(table_name, cols=[c.upper() for c in rel._field_names])
            b.select(*rel._field_refs).where(rel(*rel._field_refs)).into(table)

        # 2. execute solver job and wait for completion
        remaining_timeout_minutes = calc_remaining_timeout_minutes(
            start_time, query_timeout_mins, config_file_path=config_file_path,
        )
        print("execute solver job")
        payload = {
            "solver": solver.solver_name.lower(),
            "options": kwargs,
            "input_id": input_id,
            "data_type": self._data_type
        }
        job_id = solver._exec_job(
            payload, log_to_console=log_to_console, query_timeout_mins=remaining_timeout_minutes,
        )
        print(f"job id: {job_id}") # TODO(coey) maybe job_id is not useful

        # 3. import result relations
        print("import result relations")
        # TODO(coey) perf: only import the relations that are actually used
        to_import = [
            "result_info",
            "point",
            # "points",
            # "objective_values",
            # "primal_statuses",
        ]
        for name in to_import:
            print(f"importing {name}")
            rel = getattr(self, name)
            table_name = f"{app_name}.RAI_SOLVER.{input_id}_{name.upper()}"
            table = Table(table_name, cols=[c.upper() for c in rel._field_names])
            wheres = [getattr(table, n.upper())(r) for (n, r) in zip(rel._field_names, rel._field_refs)]
            b.define(rel(*rel._field_refs)).where(*wheres)

        print("finished solve")
        return None

    # load a particular point index from `points` into `point`
    # so it is accessible from the variable relationship
    def load_point(self, i: int):
        if not isinstance(i, int) and i >= 0:
            raise ValueError(f"Expected nonnegative integer index for point, got {i}")
        executor = self._model._to_executor()
        assert isinstance(executor, RelExecutor)
        executor.execute_raw(textwrap.dedent(f"""
        def delete[:{self.point._name}]: {self.point._name}
        def insert(:{self.point._name}, var, val): {self.points._name}(int128[{i}], var, val)
        """), readonly=False)
        return None

    # print summary of the solver result
    def summarize_result(self):
        to_get = ["error", "termination_status", "solve_time_sec", "objective_value", "solver_version", "result_count"]
        k, v = b.String.ref(), b.String.ref()
        df = b.select(k, v).where(self.result_info(k, v), k.in_(to_get)).to_df()
        print(df.to_string(index=False, header=False))
        return df

    # select variable names and values in the primal result point(s)
    def variable_values(self, multiple: bool = False):
        var = b.Hash.ref()
        val = b.Number.ref()
        s = b.String.ref()
        if multiple:
            i = b.Integer.ref()
            return b.select(i, s, val).where(self.points(i, var, val), self.variable_name(var, s))
        else:
            return b.select(s, val).where(self.point(var, val), self.variable_name(var, s))

    # get scalar information
    def __getattr__(self, name: str):
        df = None
        if name in self._model_info:
            node = b.Hash.ref()
            df = b.select(b.count(node).where(self._model_info[name](node)) | 0).to_df()
        elif name in {"error", "termination_status", "solver_version", "printed_model", "solve_time_sec", "objective_value", "result_count"}:
            val = b.String.ref()
            df = b.select(val).where(self.result_info(name, val)).to_df()
        if df is not None:
            if not df.shape == (1, 1):
                raise ValueError(f"Expected exactly one value for {name}, but df has shape {df.shape}")
            v = df.iloc[0, 0]
            if isinstance(v, str):
                if name == "solve_time_sec":
                    return float(v)
                elif name == "objective_value":
                    return int(v) if self._data_type == "int" else float(v)
                elif name == "result_count":
                    return int(v)
            return v
        return None


class ExprContext():
    def __init__(self, solver_model):
        self.solver_model = solver_model
        self.define = []
        self.where = []
        self.subcontext = []

def _rewrite(expr: b.Producer | b.Fragment, ctx: ExprContext):
    if isinstance(expr, (int, float, str)):
        return None

    elif isinstance(expr, (b.TypeRef, b.Concept)):
        return None

    elif isinstance(expr, b.Ref):
        thing = _rewrite(expr._thing, ctx)
        if thing:
            assert isinstance(thing, b.Producer)
            return thing.ref()
        return None

    elif isinstance(expr, (b.Relationship, b.RelationshipRef, b.RelationshipFieldRef)):
        rel = expr if isinstance(expr, b.Relationship) else expr._relationship
        if rel in ctx.solver_model._variable_relationships:
            return std.hash(rel._short_name, expr._parent or 0)
        return None

    elif isinstance(expr, b.Union):
        # return union of the symbolified expressions, if any are symbolic
        args_rewritten = False
        args = []
        for arg in expr._args:
            ra = _rewrite(arg, ctx)
            if ra:
                args_rewritten = True
                args.append(ra)
            else:
                args.append(arg)
        if args_rewritten:
            return b.union(*args)
        return None

    elif isinstance(expr, b.Expression):
        assert isinstance(expr._op, b.Relationship)
        assert _rewrite(expr._op, ctx) is None # shouldn't contain variables
        sym_args = [_rewrite(a, ctx) for a in expr._params]
        if all(a is None for a in sym_args):
            return None
        op = expr._op._name
        assert op in infixs or op in prefix_ops, f"Operator {op} not supported in first order solver rewrites"

        # some arguments involve solver variables, so rewrite the expression
        new_args = sym_args if op in infix_comps else sym_args[:-1] # drop last arg introduced by builder
        sm = ctx.solver_model
        node = b.Hash.ref()
        sm._expr_id += 1
        ctx.define.append(sm.operator(node, op))
        hash_args = []
        for (i, a) in enumerate(new_args):
            if a is None:
                d = sm.ordered_args_data(node, i, expr._params[i])
                hash_args.append(expr._params[i])
            else:
                d = sm.ordered_args_hash(node, i, a)
                hash_args.append(a)
            ctx.define.append(d)
        ctx.where.append(_make_hash((sm._expr_id, *hash_args), node))
        return node

    elif isinstance(expr, b.Aggregate):
        # only the last argument can be symbolic
        pre_args = expr._args[:-1]
        sym_arg = _rewrite(expr._args[-1], ctx)
        if not sym_arg:
            return None
        op = expr._op._name
        assert op in agg_ops, f"Operator {op} not supported in aggregate rewrites"

        sm = ctx.solver_model
        node = b.Hash.ref()
        ctx.where.append(_make_hash((sm._expr_id, *expr._group), node))
        # only apply where conditions in subcontext
        subctx = ExprContext(sm)
        ctx.subcontext.append(subctx)
        subctx.where.extend(expr._where._where)
        sm._expr_id += 1
        subctx.define.append(sm.operator(node, op))

        # special_ordered_set_type_2 has two ordered arguments: rank and variables
        if op == "special_ordered_set_type_2":
            assert len(pre_args) == 1, "special_ordered_set_type_2 expects exactly 2 arguments (rank, variables)"
            subctx.define.append(sm.ordered_args_hash(node, pre_args[0], sym_arg))
        else:
            # other aggregate operators use unordered args
            arg_hash = b.Hash.ref()
            subctx.where.append(_make_hash((sm._expr_id, *pre_args), arg_hash))
            subctx.define.append(sm.unordered_args_hash(node, arg_hash, sym_arg))
        return node

    elif isinstance(expr, b.Fragment):
        # only support selects with one item
        assert not expr._define and not expr._require and len(expr._select) == 1
        sym_select = _rewrite(expr._select[0], ctx)
        if sym_select:
            return b.select(sym_select).where(*expr._where)
        return None

    raise NotImplementedError(f"Solver rewrites cannot handle {expr} of type {type(expr)}")

def _expr_strings_rec(x, names_dict, ops_dict, args_dict):
    if x in names_dict:
        # this is a variable node
        return names_dict[x]
    elif x not in ops_dict:
        # this is a value
        return str(x)

    # this is an expression
    assert x in ops_dict
    arg_strs = []
    op = ops_dict[x]

    for k in args_dict.get(x, []):
        s = _expr_strings_rec(k, names_dict, ops_dict, args_dict)
        if op in infix_ops and k in ops_dict and ops_dict[k] in infixs:
            # add parentheses to avoid precedence issues
            s = f"({s})"
        arg_strs.append(s)

    if op in agg_ops and not op == "special_ordered_set_type_2":
        # sort unordered args to improve determinism
        arg_strs.sort()

    if op in infixs:
        # infix operator, e.g. x + y
        assert len(arg_strs) == 2 # TODO could relax e.g. for +, *
        return f"{arg_strs[0]} {op} {arg_strs[1]}"
    else:
        # prefix operator, e.g. abs(x)
        assert len(arg_strs) >= 1
        return f"{op}({', '.join(arg_strs)})"

def _make_hash(tup, res):
    return b.Expression(b.Relationship.builtins["hash"], b.TupleArg(tup), res)

infix_ops = set(["+", "-", "*", "/", "^"])
infix_comps = set(["=", "!=", "<", "<=", ">", ">=", "implies"])
infixs = infix_ops.union(infix_comps)
prefix_ops = set(["abs", "log", "exp"])
agg_ops = set(["sum", "count", "min", "max", "all_different", "special_ordered_set_type_2"])

# _variable_types = {
#     "cont": 40,
#     "int": 41,
#     "bin": 42,
# }

# _fo_operators = {
#     "+": 10,
#     "-": 11,
#     "*": 12,
#     "/": 13,
#     "^": 14,
#     "abs": 20,
#     "exp": 21,
#     "log": 22,
#     "range": 50,
# }

# _fo_comparisons = {
#     "=": 30,
#     "!=": 31,
#     "<=": 32,
#     ">=": 33,
#     "<": 34,
#     ">": 35,
#     "implies": 62,
# }

# _ho_operators = {
#     "sum": 80,
#     # "product":81,
#     "min": 82,
#     "max": 83,
#     "count": 84,
#     "all_different": 90,
# }



import warnings

from v0.relationalai.semantics.metamodel.builtins import concept_relation_annotation, from_cdc_annotation

__all__ = ["concept_relation_annotation", "from_cdc_annotation"]

warnings.warn(
    "relationalai.early_access.metamodel.builtins is deprecated, "
    "Please migrate to relationalai.semantics.metamodel.builtins",
    DeprecationWarning,
    stacklevel=2,
)


from v0.relationalai.experimental.paths.api import path, union
from v0.relationalai.semantics import Concept

Node = Concept('Node')
OddNode = Concept('OddNode')
EvenNode = Concept('EvenNode')

# pattern = path(node(Node, lambda n: n.id > 3), star('-[A]->', '-[B]->'), Node)

# pattern = path(node(Node, lambda n: n.id > 3), Node)

# pattern = path(Node, '-[A]->', Node, '-[B]->', Node)

# pattern = path('-[B]->', node(Node, lambda n: n.id > 3), '-[A]->')

# pattern = path(OddNode, star('-[A]->', '-[B]->'), EvenNode)

# pattern = path(OddNode, star('-[A]->', node(OddNode, lambda n: n.id > 3)), EvenNode)

pattern = path(union("-[A]->", "-[B]->"), union("-[A]->", "-[B]->"))

g = pattern.glushkov()
a = g.automaton()
a.reduce()

print(a)





import warnings

from v0.relationalai.semantics.tests.test_snapshot_base import SnapshotTestBase

__all__ = ["SnapshotTestBase"]

warnings.warn(
    "relationalai.early_access.tests.test_snapshot_base is deprecated. "
    "Please migrate to relationalai.semantics.tests.test_snapshot_base",
    DeprecationWarning,
    stacklevel=2,
)


"""
Example usage of automaton-based pathfinding.
"""

import time
from v0.relationalai.semantics import Model, Integer, select, String, define
from v0.relationalai.experimental.paths.graph import Graph
from v0.relationalai.experimental.paths.api import node, path, match

import argparse


# def print_paths(path_node, path_label):
#     # Print nodes and labels in a readable format
#     l2 = String.ref()
#     k = Integer.ref()
#     m = Integer.ref()
#     print("\nNodes and labels in the path on the original graph:")
#     select(k, m, l2).where(
#         path_node(k, m),
#         path_label(k, label := String.ref()),
#         l2 == strings.concat("-[", label, "]->")
#     ).inspect()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Pathfinder example script.")
    parser.add_argument("--num_paths", type=int, default=None, help="Number of paths to find.")
    args = parser.parse_args()

    print("Number of paths to find:", args.num_paths)

    # Create an automaton that matches paths with pattern: A(A*)B:
    # pattern = path("-[A]->", star("-[A]->"), "-[B]->")
    # edge_list = [(1, 2, "A"), (2, 3, "A"), (3, 4, "A"), (4, 5, "B"), (1, 5, "A")]

    # 2 paths matching AB:
    # pattern = path("-[A]->", "-[B]->")
    # edge_list = [(1, 2, "A"), (2, 3, "B"), (1, 4, "A"), (4, 3, "B")]

    

    # pattern = path("-[A]->")
    # edge_list = [(1, 3, "A"), (2, 3, "A")]

    pattern = path("A")
    # edge_list = [(1, 3, "A"), (2, 3, "A"), (1, 4, "A"), (2, 4, "A")]
    edge_list = [(1, 2, "A")]


    sources = []
    targets = []

    sources = [1]
    targets = [2]

    # pattern = path("-[A]->")
    # edge_list = [(1, 3, "A"), (2, 3, "A"), (1, 4, "A"), (2, 4, "A")]

    # sources = [1, 2]
    # targets = [3, 4]
    # 2 paths from 2 different source nodes:
    # pattern = path("-[A]->")
    # edge_list = [(1, 3, "A"), (2, 3, "A")]

    # pattern = path("-[A]->")
    # edge_list = [(1, 3, "A"), (2, 3, "A"), (1, 4, "A"), (2, 4, "A")]


    print("\nEdge list:\n", edge_list)

    # Initialize the RelationalAI model
    model = Model("paths", use_lqp=True) # , use_lqp=True)  # strict=True  # dry_run=False  # use_direct_access=True
    # use_lqp=True  # strict=True  # dry_run=False  # use_direct_access=True

    # model.define(
    #     RawSource("rel", "@span_no_threshold")
    # )

    graph = Graph.from_edge_list(model, edge_list)
    Node = graph.Node

    # Call the function to find the path

    Source = model.Relationship(f"source5 {{{Node}}}")
    Target = model.Relationship(f"target5 {{{Node}}}")

    for src in sources:
        define(Source(Node.new(id=src)))

    for tgt in targets:
        define(Target(Node.new(id=tgt)))

    src, tgt = Node.ref(), Node.ref()

    # select(graph.Node).inspect()

    graph.Edge.inspect()
    select(src).where(Source(src)).inspect()
    select(tgt).where(Target(tgt)).inspect()

    # print(f"\nFinding paths matching the automaton pattern from {src} to {dst}..."
    # projected_paths = match(graph, pattern, Source, Target, num_paths=args.num_paths)

    start = time.time()
    projected_paths = match(graph, pattern, Source, Target)

    i, k = Integer.ref(), Integer.ref()
    node = graph.Node.ref()
    label = String.ref()

    result = select(i, k, label, node.id).where(projected_paths(i, k, label, node)).inspect()
    end = time.time()

    print(f"Time taken to find paths: {end - start} seconds")
